#from __future__ import (absolute_import, division,
#                        print_function, unicode_literals)
#from builtins import *

import sys
import os
import itertools
import numpy as np

from datautils.xrayutils import id31_tools_5 as id31

import fabio
import threading
import PyMca5.PyMcaIO.PilatusCBF as cbf
from silx.io.specfile import SpecFile
import time

#from util.id31_tools_5 import P3_Image, Fastscan
from datautils.xrayutils import HKLVlieg
from datautils.xrayutils import DetectorCalibration

from .. import backend, errors, util

try:
    from PyMca import SixCircle
except ImportError:
    from PyMca5.PyMca import SixCircle



# TODO:
# - switch to parallel handling of images instead of scans
# - implement Willem rotations
# - implement UBcalc for HKLProjection
# - run PBCO files

########################################################################

class HKLProjection(backend.ProjectionBase):
    # arrays: gamma, delta
    # scalars: theta, mu, chi, phi
    def project(self, k0, ub, gamma, delta, theta, mu, chi,
                phi):

        wavelength = id31.get_lambda(k0)
        R = SixCircle.getHKL(wavelength, ub, gamma=gamma - mu, delta=delta,
                             theta=theta, mu=mu, chi=chi, phi=phi)

        shape = gamma.size, delta.size
        H = R[0, :].reshape(shape)
        K = R[1, :].reshape(shape)
        L = R[2, :].reshape(shape)
        
        return (H, K, L)

    def get_axis_labels(self):
        return 'H', 'K', 'L'
    
class HKLProj_AutoLimits(backend.ProjectionBase):
    def parse_config(self, config):
        super(backend.ProjectionBase, self).parse_config(config)
        res = config.pop('resolution')  # or just give 1 number for all dimensions
        self.config.auto_limits = util.parse_pairs(config.pop('auto_limits', None))
        self.config.limits = util.parse_pairs(config.pop('limits', None))  # Optional, set the limits of the space object in projected coordinates. Syntax is same as numpy e.g. 'limits = [:0,-1:,:], [0:,:-1,:], [:0,:-1,:], [0:,-1:,:]'
        print("config: " + str(config))
        command = config.pop('command')#util.parse_multi_range(','.join(command).replace(' ',','))
        scans = command[0].split(',')
        
        self.config.rod = np.array(command[1:],dtype=np.float)
        if self.config.auto_limits:
            if not self.config.limits:
                self.config.limits = []
            for lim in self.config.auto_limits:
                Hlim = slice(lim[0].start + self.config.rod[0],
                             lim[0].stop + self.config.rod[0])
                Klim = slice(lim[1].start + self.config.rod[1],
                             lim[1].stop + self.config.rod[1])
                self.config.limits.append([Hlim,Klim,lim[2]])
            
            #self.config.limits = self.config.auto_limits
            #print(self.config.limits[0][0])
            
            
        print("limits = " + str(self.config.limits))
        
        labels = self.get_axis_labels()
        if not self.config.limits is None:
            for lim in self.config.limits:
                if len(lim) != len(labels):
                    raise errors.ConfigError('dimension mismatch between projection axes ({0}) and limits specification ({1}) in {2}'.format(labels, self.config.limits, self.__class__.__name__))
        if ',' in res:
            self.config.resolution = util.parse_tuple(res, type=float)
            if not len(labels) == len(self.config.resolution):
                raise errors.ConfigError('dimension mismatch between projection axes ({0}) and resolution specification ({1}) in {2}'.format(labels, self.config.resolution, self.__class__.__name__))
        else:
            self.config.resolution = tuple([float(res)] * len(labels))
        
    # arrays: gamma, delta 
    # scalars: theta, mu, chi, phi
    def project(self, VliegAngles, gamma_cryst, delta, om, mu_cryst, chi, phi):
        H,K,L = VliegAngles.anglesToHklDetector_mesh(mu_cryst,delta,gamma_cryst,om,chi,phi)
        #print(H.shape)
        print(np.mean(H),np.mean(K),np.mean(L))
        return (H, K, L)

    def get_axis_labels(self):
        return 'H', 'K', 'L'


class QProjection(backend.ProjectionBase):
    def project(self, k0, ub, gamma, delta, theta, mu, chi, phi):
        del_grid = np.tile(delta, (np.size(gamma), 1))
        gam_grid = np.transpose(np.tile(gamma, (np.size(delta), 1)))

        qx, qy, qz = id31.get_q(k0, theta, del_grid, gam_grid)

        return (qx.flatten(), qy.flatten(), qz.flatten()) 

    def get_axis_labels(self):
        return 'qx', 'qy', 'qz'
        
class QSIXCProjection(backend.ProjectionBase):
    def project(self, k0, UB, gamma, delta, theta, mu, chi, phi):
        wavelength = id31.get_lambda(k0)
        print(wavelength)
        shape = gamma.size, delta.size
        sixc = SixCircle.SixCircle()
        sixc.setLambda(wavelength)
        sixc.setUB(UB)
        R = sixc.getQSurface(gamma=gamma-mu, delta=delta, theta=theta, mu=mu, chi=chi, phi=phi)
        qz = R[0, :].reshape(shape)
        qy = R[1, :].reshape(shape)
        qx = R[2, :].reshape(shape)
        return (qz, qy, qx)

    def get_axis_labels(self):
        return 'qx', 'qy', 'qz'

########################################################################


class EH1(backend.InputBase):
    def generate_jobs(self, command):
        print("generate_jobs")
        scans = command[0].split(',')
        
        self.config.rod = np.array(command[1:],dtype=np.float)
        cal_xy = True        
        
        if self.config.rod.size != 2:
            raise NotImplementedError("Extraction of multiple rods at the same time is not implemented jet ")
        
        
        for scan in scans:
            
            
            # this is for manual calcualtion of the angles          
            
            if cal_xy:            
                hkl = self.config.cal_reflections[:3].T
                xy = self.config.cal_reflections[3:5].T
                omega = np.deg2rad(self.config.cal_reflections[5])*-1
                
                gamma, delta,mu_cryst = self.config.dc.crystalAnglesPoint(xy.T[1],xy.T[0],self.config.mu,self.config.refractionindex)
                alpha = np.full(delta.shape,mu_cryst)
                chi = np.zeros(delta.shape)
                phi = np.zeros(delta.shape)
                positions = np.array([alpha,delta,gamma,omega,chi,phi]).T
            else:
                hkl = self.config.cal_reflections[:3].T
                positions = np.deg2rad(self.config.cal_reflections[3:])
                #positions[2] -= positions[0] #  pos[0] = mu
                positions[3] *= -1. # (om is lefthanded)
                positions = positions.T

            #print(np.rad2deg(positions))
            
            
            """
            dynamic calculation
            """
            
            #positions = HKLVlieg.crystalAngles(positions,self.config.refractionindex)
            #mu_cryst = HKLVlieg.crystalAngles_singleArray(self.config.mu,self.config.refractionindex)
            
            # calculation of UB        
            i = 1
            j = 7
            self.config.UB.setPrimaryReflection(positions[i],hkl[i])
            self.config.UB.setSecondayReflection(positions[j],hkl[j])
            self.config.UB.calculateU()
            self.config.UB.refineU(hkl,positions,False,self.config.rod,4.)
            
            
            #reflections_rod = np.all(hkl[:,:2] == self.config.rod,axis=1)
            
            #print(self.config.xmirror)
            
            self.config.VliegAngles = HKLVlieg.VliegAngles(self.config.UB)
            #self.config.gc = HKLVlieg.GeometryCorrection(self.config.VliegAngles)
            if self.config.isBliss:
                fastscan = id31.BlissScan(self.config.spec,scan)
            else:
                fastscan = id31.Fastscan(self.config.spec,int(scan))
            fastscan.set_image_folder(self.config.img_prefix)
            fastscan.set_th_offset(self.config.th_offset)
            

            omegas = np.deg2rad(fastscan.omega)
    
            posmax = self.config.VliegAngles.anglesZmode_np18(np.concatenate((self.config.rod , [self.config.L_range[1]])),\
                                                mu_cryst,'in',mirrorx=False)
            posmin = self.config.VliegAngles.anglesZmode_np18(np.concatenate((self.config.rod ,[self.config.L_range[0]])),\
                                                mu_cryst,'in',mirrorx=False)
            posmax_m = self.config.VliegAngles.anglesZmode_np18(np.concatenate((self.config.rod , [self.config.L_range[1]])),\
                                                mu_cryst,'in',mirrorx=True)
            posmin_m = self.config.VliegAngles.anglesZmode_np18(np.concatenate((self.config.rod ,[self.config.L_range[0]])),\
                                                mu_cryst,'in',mirrorx=True)

            """
            if 'xmirror' not in self.config.__dict__:
                if np.any(reflections_rod):
                    examplepos = positions[reflections_rod][0]
                    self.config.xmirror = examplepos[1] < 0.
                else:
                    raise Exception("Can not determine xmirror. Please include xmirror to the config!")
            """
            print("rod = (%.2f ,%.2f )" % tuple(self.config.rod))
            print("omrange  = %.2f - %.2f" % (np.rad2deg(omegas[0]),np.rad2deg(omegas[-1])))
            
            if 'xmirror' not in self.config.__dict__:

                indxlowl = np.argmin(np.abs(omegas - posmin[3]))
                indxhighl = np.argmin(np.abs(omegas - posmax[3]))
    
                indxlowl_m = np.argmin(np.abs(omegas - posmin_m[3]))
                indxhighl_m = np.argmin(np.abs(omegas - posmax_m[3]))
                
                noimages = np.abs(indxlowl - indxhighl) 
                noimages_m = np.abs(indxlowl_m - indxhighl_m)
                
                print("not mirrored = %.2f - %.2f, %04i - %04i" % (np.rad2deg(posmin[3]),np.rad2deg(posmax[3]),indxlowl,indxhighl))      
                print("mirrored = %.2f - %.2f, %04i - %04i" % (np.rad2deg(posmin_m[3]),np.rad2deg(posmax_m[3]),indxlowl_m,indxhighl_m)) 
                if noimages >= omegas.size-1:
                    noimages = 0
                if noimages_m >= omegas.size-1:
                    noimages_m = 0
                
                self.config.xmirror = noimages_m > noimages
                
            
            
            print("Decide to xmirror = %s" % self.config.xmirror)
            print("Continue in ...")
            for i in range(3,0,-1):
                print(i)
                time.sleep(1)
                
            
            if self.config.xmirror:
                self.config.ommax = posmax_m[3]
                self.config.ommin = posmin_m[3]
            else:
                self.config.ommax = posmax[3]
                self.config.ommin = posmin[3]
                
            indxlowl = np.argmin(np.abs(omegas - self.config.ommin))
            indxhighl = np.argmin(np.abs(omegas - self.config.ommax))
            
            if indxlowl < indxhighl:
                indxlow = indxlowl -15
                indxhigh = indxhighl 
            else:
                indxlow = indxhighl
                indxhigh = indxlowl +15
            
            if indxlow < 0:
                indxlow = 0
            if indxhigh >= omegas.size:
                indxhigh = omegas.size
                
            for imageno in range(indxlow,indxhigh):
                yield backend.Job(scanno = scan, imageno = imageno)
        
        #ths,filenames = self.getParam(str(self.config.spec),scan[0])
        #print(scan)

        #for th,filename in zip(ths,filenames):
        

    ####################################################################

    def process_job(self, job):
        ################################################################

        super(EH1, self).process_job(job)
        scan = job.scanno
        imageno = job.imageno
        
        if self.config.isBliss:
            fastscan = id31.BlissScan(self.config.spec,scan)
        else:
            fastscan = id31.Fastscan(self.config.spec,int(scan))
        fastscan.set_image_folder(self.config.img_prefix)
        #fastscan.estimateRingCurrent(self.config.lossps)
        fastscan.set_th_offset(self.config.th_offset)
        
        image = fastscan[imageno]
        ysize, xsize = image.img.shape
        #y = np.arange(ysize)
        #x = np.arange(xsize)
        #delta, gamma, detcorrection = self.config.dc.xyToDelGam_corr(x,y)
        #gamma -= self.config.mu
        #self.config.dc.setAzimuthalReference( self.config.azimuth)
        #gamma, delta = self.config.dc.surfaceAngles(self.config.mu)
        #transform into crystal frame
        #delta = self.config.delta
        #gamma_cryst = self.config.gamma_cryst #HKLVlieg.crystalAngles_singleArray(gamma,self.config.refractionindex)
        #mu_cryst = self.config.mu_cryst # HKLVlieg.crystalAngles_singleArray(self.config.mu,self.config.refractionindex)
        
        #ddel, ggam = np.meshgrid(delta,gamma_cryst)
        
        #air absorption makes no difference!!!(0.5 % I error max, but rather less!!!)
        #sgamdel = np.arctan(np.sqrt( np.tan(ggam)**2 + np.tan(ddel)**2 ))        
        
        #air_distance = self.config.distance/np.cos(sgamdel)
        
        #C_air = np.exp(- self.config.air_absorbtion * air_distance)
        
        #C_det_pol = self.config.dc.correctionArray(-1.)
        
        self.config.dc.setAzimuthalReference( self.config.azimuth)
        self.config.dc.setPolarization(self.config.polaxis,self.config.polfactor)
        gamma_cryst,delta,mu_cryst = self.config.dc.crystalAngles(self.config.mu,self.config.refractionindex)
        chi = 0.0
        phi = 0.0
        image.img /= self.config.dc.correctionArray()
        #image.img *= (80.0/image.counters['srcur'])
        image.img *= (1.0/image.counters['TrigTime'])
        #image.img /= C_pol
        #image.img /= C_air
        
        weights = np.ones_like(image.img)
        
        # Masking
        if 'mask' in self.config.__dict__:
            if self.config.mask.shape != image.img.shape:
                raise errors.BackendError('The mask matrix does not have the same shape as the images')
            else:
                weights[self.config.mask] = 0
                image.img[self.config.mask] = 0
        else:
            print('Mask is not used')
        omega = np.deg2rad(image.counters['om'])
        intensity = image.img
        

        
        h,k,l = self.config.VliegAngles.anglesToHklDetector_mesh(mu_cryst,delta,gamma_cryst,omega,0.,0.)

        l_low = l < 0.
        lmask = np.zeros_like(l)
        lmask[l_low] = 100.
        weightfun = np.abs(h - self.config.rod[0]) + np.abs(k - self.config.rod[1] ) + lmask
   
        ind = np.unravel_index(np.argmin(weightfun, axis=None), h.shape)
        print("pos of rod: %s,%s" % ind)
        gam_c = gamma_cryst[ind]
        
        roiheight_c = int(self.config.roiheight_start - np.rad2deg(gam_c) * self.config.roiheight_rate)
        if roiheight_c < self.config.roiheight_min:
            roiheight_c = self.config.roiheight_min
        
        heightrange =  np.array((int(ind[0] - roiheight_c/2 + self.config.roiheightoffset) , int(ind[0] + roiheight_c/2 + self.config.roiheightoffset)),dtype=np.int)
        widthrange = np.array((int(ind[1] - self.config.roiwidth/2 + self.config.roiwidthoffset) , int(ind[1] + self.config.roiwidth/2 + self.config.roiwidthoffset)),dtype=np.int)

        indxgamzero = np.argmin(np.abs(gamma_cryst[:,ind[1]]))
        if np.abs(indxgamzero - ind[0]) < self.config.roiheight_start*2:
            heightrange[1] = indxgamzero + 10
        
        heightrange = np.clip(heightrange,0,ysize)
        widthrange = np.clip(widthrange,0,xsize)
        
        #croi = intensity[ heightrange[0]: heightrange[1],widthrange[0]:widthrange[1]]
        
        completeheightrange = heightrange
        completewidthrange = widthrange
        completewidthrange[0] -= self.config.roiwidth
        completewidthrange[0] += self.config.roiwidth
        
        completeheightrange = np.clip(heightrange,0,ysize)
        completewidthrange = np.clip(completewidthrange,0,xsize)
        
        #print(heightrange[0], heightrange[1],widthrange[0]+self.config.roiwidth,int(widthrange[1]+self.config.roiwidth))
        bgroi1 = intensity[ heightrange[0]: heightrange[1],int(widthrange[0]+self.config.roiwidth):int(widthrange[1]+self.config.roiwidth)]
        bgroi2 = intensity[ heightrange[0]: heightrange[1],int(widthrange[0]-self.config.roiwidth):int(widthrange[1]-self.config.roiwidth)]
        
        bgroi1ct = weights[ heightrange[0]: heightrange[1],int(widthrange[0]+self.config.roiwidth):int(widthrange[1]+self.config.roiwidth)]
        bgroi2ct = weights[ heightrange[0]: heightrange[1],int(widthrange[0]-self.config.roiwidth):int(widthrange[1]-self.config.roiwidth)]
        
        bg = (np.nansum(bgroi1) + np.nansum(bgroi2)) / (np.nansum(bgroi1ct) + np.nansum(bgroi2ct))
        
        completeroi = intensity[ completeheightrange[0]: completeheightrange[1],completewidthrange[0]:completewidthrange[1]]
        #completeroi -= bg
        
        #croi -= bg
        #croi += 10.
        
        
        #weights = weights[ heightrange[0]: heightrange[1],widthrange[0]:widthrange[1]]
        weights = weights[ completeheightrange[0]: completeheightrange[1],completewidthrange[0]:completewidthrange[1]]
        gamma_cryst = gamma_cryst[ completeheightrange[0]: completeheightrange[1],completewidthrange[0]:completewidthrange[1]]
        delta = delta[ completeheightrange[0]: completeheightrange[1],completewidthrange[0]:completewidthrange[1]]
        # Cropping
        #gamma_grid = apply_crop(gamma_cryst, self.config.range_x, self.config.range_y)
        #delta_grid = apply_crop(delta, self.config.range_x, self.config.range_y)
        #gamma_cryst = gamma_cryst[self.config.range_y]
        #delta = delta[self.config.range_x]
        #weights = apply_crop(weights, self.config.range_x, self.config.range_y)
        #intensity = apply_crop(image.img, self.config.range_x, self.config.range_y)
        #print(gamma_cryst.shape)
        #print(delta.shape)
        #print(intensity.shape)
        
        om = omega
        
        #print('Processing file ' + str(filename))
        print("om = %.2f" % (np.rad2deg(om)))
        
        

        yield completeroi, weights, (self.config.VliegAngles, gamma_cryst, delta,
                                       om, mu_cryst, chi, phi)

    ####################################################################

    def parse_config(self, config):
        super(EH1, self).parse_config(config)
        print("config")

        self.config.img_prefix = str(config.pop('img_prefix'))
        #self.config.cncx_offset = float(config.pop('cncx_offset'))
        #self.config.cen_x = float(config.pop('cen_x'))
        #self.config.cen_y = float(config.pop('cen_y'))
        #self.config.px_size_mm = float(config.pop('px_size_mm'))
        #self.config.E_keV = float(config.pop('e_kev'))
        self.config.mu = np.deg2rad( float(config.pop('mu')))
        self.config.azimuth = np.deg2rad( float(config.pop('azimuth')))
        self.config.polaxis = np.deg2rad( float(config.pop('polaxis')))
        self.config.polfactor = np.deg2rad( float(config.pop('polfactor')))
        
        self.config.refractionindex = float(config.pop('refractionindex'))
        #self.config.range_x = util.parse_multi_range(config.pop('range_x', None))
        #self.config.range_y = util.parse_multi_range(config.pop('range_y', None))
        self.config.isBliss = util.parse_bool( config.pop('isbliss'))
        
        ponifile = config.pop('det_cal')
        dc = DetectorCalibration.Detector2D_SXRD()
        dc.load(ponifile)
        dc.setAzimuthalReference( self.config.azimuth)
        dc.setPolarization(self.config.polaxis,self.config.polfactor)
        gamma_cryst,delta,mu_cryst = dc.crystalAngles(self.config.mu,self.config.refractionindex)
        self.config.dc = dc
        
        
        
        #self.config.dc.setCalibration([self.config.cen_x,self.config.cen_y],self.config.distance,self.config.detrot)
        mask_path = config.pop('mask')
                # Masking
        if os.path.isfile(mask_path):
            self.config.mask = np.load(mask_path).astype(np.bool)
        
        self.config.spec = config.pop('spec_file')
        #self.config.distance = float(config.pop('distance'))
        #self.config.detrot = float(config.pop('detrot'))
        
        #self.config.air_absorbtion = float(config.pop('air_abs'))
        #self.config.lossps = float(config.pop('lossps'))        
        
        #ub = config.pop('UB')
        crystalparam = np.array(util.parse_tuple(config.pop('crystal'), length=6, type=float))
        self.config.crystal = HKLVlieg.Crystal(crystalparam[:3],crystalparam[3:]) 
        self.config.L_range = np.array(util.parse_tuple(config.pop('l_range'), length=2, type=float))
        

        self.config.UB = HKLVlieg.UBCalculator(self.config.crystal,1.)
        self.config.UB.setLambda(self.config.dc.get_wavelength()*1e10)
        self.config.E_keV = self.config.UB.getEnergy()
        

        #self.config.UB.setU(np.array(util.parse_tuple(config.pop('orient'), length=9, type=float)))
        
        
        
        self.config.cal_reflections = np.loadtxt(config.pop('cal_reflections'),comments='#').T
        if 'xmirror' in config:
            self.config.xmirror = util.parse_bool(config.pop('xmirror'))
        
        if 'th_offset' in config:
            self.config.th_offset = float(config.pop('th_offset'))
        else:
            self.config.th_offset = 0.0
        
        self.config.roiheight_start = float(config.pop('roiheight_start'))
        self.config.roiheight_min = float(config.pop('roiheight_min'))
        self.config.roiheight_rate = float(config.pop('roiheight_rate'))
        self.config.roiheightoffset = float(config.pop('roiheightoffset'))
        
        self.config.roiwidth = float(config.pop('roiwidth'))
        self.config.roiwidthoffset = float(config.pop('roiwidthoffset'))
        
        #gamma, delta = self.config.dc.surfaceAngles(self.config.mu)
        #transform into crystal frame
        #gamma_cryst = HKLVlieg.crystalAngles_singleArray(gamma,self.config.refractionindex)
        #self.config.mu_cryst = HKLVlieg.crystalAngles_singleArray(self.config.mu,self.config.refractionindex)
        
        #ddel, ggam = np.meshgrid(delta,gamma_cryst)
        
        #air absorption makes no difference!!!(0.5 % I error max, but rather less!!!)
        #sgamdel = np.arctan(np.sqrt( np.tan(ggam)**2 + np.tan(ddel)**2 ))        
        
        #air_distance = self.config.distance/np.cos(sgamdel)
        
        #C_air = np.exp(- self.config.air_absorbtion * air_distance)
        
        #C_det_pol = self.config.dc.correctionArray()
        
        #self.config.gamma_cryst = gamma_cryst
        #self.config.gamma = gamma
        #self.config.delta = delta


    ####################################################################

    def get_destination_options(self, command):
        if not command:
            return False
        scans = command[0].split(',')
        if len(scans) > 1:    
            scan = "%s-%s" % (scans[0],scans[-1])
        else:
            scan = scans[0]
        
        
        rod = np.array(command[1:],dtype=np.float)
        H = "%i_%02i" % (round(rod[0]), round((rod[0] % 1) * 100))
        K = "%i_%02i" % (round(rod[1]), round((rod[1] % 1) * 100))
        
        return dict(scanno=scan,
                    H=H,
                    K=K)

   
class EH1_fast_complete(backend.InputBase):
    def generate_jobs(self, command):
        print("generate_jobs")
        command = np.array(command,dtype=np.float)#util.parse_multi_range(','.join(command).replace(' ',','))
        scans = [int(command[0])]
        cal_xy = False
        
        #self.config.rod = command[1:]
        #if self.config.rod.size != 2:
        #    raise NotImplementedError("Extraction of multiple rods at the same time is not implemented jet ")
        
        for scan in scans:
    
            if cal_xy:            
                hkl = self.config.cal_reflections[:3].T
                xy = self.config.cal_reflections[3:5].T
                omega = np.deg2rad(self.config.cal_reflections[5])*-1
                
                delta, gamma = self.config.dc.xyToDelGam(xy.T[0],xy.T[1])
                gamma = gamma - self.config.mu
                
                alpha = np.full(delta.shape,self.config.mu)
                chi = np.zeros(delta.shape)
                phi = np.zeros(delta.shape)
                positions = np.array([alpha,delta,gamma,omega,chi,phi]).T                    
            else:
                hkl = self.config.cal_reflections[:3].T
                positions = np.deg2rad(self.config.cal_reflections[3:])
                positions[2] -= positions[0] # = mu
                positions[3] *= -1. # (om is lefthanded)

            
            """
            dynamic calculation
            """
            
            positions = HKLVlieg.crystalAngles(positions,self.config.refractionindex)
            #mu_cryst = HKLVlieg.crystalAngles_singleArray(self.config.mu,self.config.refractionindex)
            
            # calculation of UB        
            i = 1
            j = 2
            self.config.UB.setPrimaryReflection(positions[i],hkl[i])
            self.config.UB.setSecondayReflection(positions[j],hkl[j])
            self.config.UB.calculateU()
            self.config.UB.refineU(hkl,positions,False)
            
            
            
            #reflections_rod = np.all(hkl[:,:2] == self.config.rod,axis=1)
            
            #print(self.config.xmirror)
            
            self.config.VliegAngles = HKLVlieg.VliegAngles(self.config.UB)
            self.config.gc = HKLVlieg.GeometryCorrection(self.config.VliegAngles)
            
            if self.config.isBliss:
                fastscan = id31.BlissScan(self.config.spec,scan)
            else:
                fastscan = id31.Fastscan(self.config.spec,int(scan))
            fastscan.set_image_folder(self.config.img_prefix)
            fastscan.set_th_offset(self.config.th_offset)
            
                
            for imageno in range(fastscan.nopoints):
                yield backend.Job(scanno = scan, imageno = imageno)
        
        #ths,filenames = self.getParam(str(self.config.spec),scan[0])
        #print(scan)

        #for th,filename in zip(ths,filenames):
        

    ####################################################################

    def process_job(self, job):

        def apply_crop(data, xmask, ymask):
            roi = data[ymask, :]
            return roi[:, xmask]

        ################################################################

        super(EH1_fast_complete, self).process_job(job)
        scan = job.scanno
        imageno = job.imageno
        if self.config.isBliss:
            fastscan = id31.BlissScan(self.config.spec,scan)
        else:
            fastscan = id31.Fastscan(self.config.spec,int(scan))
        fastscan.set_image_folder(self.config.img_prefix)
        fastscan.estimateRingCurrent(self.config.lossps)
        fastscan.set_th_offset(self.config.th_offset)
        
        image = fastscan[imageno]
        ysize, xsize = image.img.shape
        y = np.arange(ysize)
        x = np.arange(xsize)
        delta, gamma, detcorrection = self.config.dc.xyToDelGam_corr(x,y)
        gamma -= self.config.mu
             
        
        #transform into crystal frame
        gamma_cryst = HKLVlieg.crystalAngles_singleArray(gamma,self.config.refractionindex)
        mu_cryst = HKLVlieg.crystalAngles_singleArray(self.config.mu,self.config.refractionindex)
        
        ddel, ggam = np.meshgrid(delta,gamma_cryst)
        
        #air absorption makes no difference!!!(0.5 % I error max, but rather less!!!)
        #sgamdel = np.arctan(np.sqrt( np.tan(ggam)**2 + np.tan(ddel)**2 ))        
        
        #air_distance = self.config.distance/np.cos(sgamdel)
        
        #C_air = np.exp(- self.config.air_absorbtion * air_distance)
        
        C_pol = self.config.gc.polarization(ddel,ggam,mu_cryst,0.)
        
        chi = 0.0
        phi = 0.0
        image.img /= detcorrection
        image.img *= (80.0/image.counters['srcur'])
        image.img *= (1.0/image.counters['TrigTime'])
        image.img /= C_pol
        #image.img /= C_air
        
        weights = np.ones_like(image.img)
        
        # Masking
        if os.path.isfile(self.config.mask):
            #mask_img = fabio.open(self.config.mask).data
            mask_img = np.load(self.config.mask).astype(np.bool)
            if mask_img.shape != image.img.shape:
                raise errors.BackendError('The mask matrix does not have the same shape as the images')
            else:
                weights[mask_img] = 0
                image.img[mask_img] = 0
        else:
            print('Mask is not used')
        """
        import matplotlib.pyplot as plt
        plt.imshow(C_air)
        plt.show()
        """
        omega = np.deg2rad(image.counters['om'])
        intensity = image.img
        
        completeroi = intensity
        #completeroi -= bg
        
        #croi -= bg
        #croi += 10.
        
        
        #weights = weights[ heightrange[0]: heightrange[1],widthrange[0]:widthrange[1]]
        weights = weights
        gamma_cryst = gamma_cryst
        delta = delta
        # Cropping
        #gamma_grid = apply_crop(gamma_cryst, self.config.range_x, self.config.range_y)
        #delta_grid = apply_crop(delta, self.config.range_x, self.config.range_y)
        #gamma_cryst = gamma_cryst[self.config.range_y]
        #delta = delta[self.config.range_x]
        #weights = apply_crop(weights, self.config.range_x, self.config.range_y)
        #intensity = apply_crop(image.img, self.config.range_x, self.config.range_y)
        #print(gamma_cryst.shape)
        #print(delta.shape)
        #print(intensity.shape)
        
        om = omega
        
        #print('Processing file ' + str(filename))
        print("om = %.2f" % (np.rad2deg(om)))
        
        

        yield completeroi, weights, (self.config.VliegAngles, gamma_cryst, delta,
                                       om, mu_cryst, chi, phi)

    ####################################################################

    def parse_config(self, config):
        super(EH1_fast_complete, self).parse_config(config)
        print("config")

        self.config.img_prefix = str(config.pop('img_prefix'))
        #self.config.cncx_offset = float(config.pop('cncx_offset'))
        self.config.cen_x = float(config.pop('cen_x'))
        self.config.cen_y = float(config.pop('cen_y'))
        self.config.px_size_mm = float(config.pop('px_size_mm'))
        self.config.E_keV = float(config.pop('e_kev'))
        self.config.mu = np.deg2rad( float(config.pop('mu')))
        #self.config.range_x = util.parse_multi_range(config.pop('range_x', None))
        #self.config.range_y = util.parse_multi_range(config.pop('range_y', None))
        
        
        
        self.config.mask = config.pop('mask')
        self.config.spec = config.pop('spec_file')
        self.config.distance = float(config.pop('distance'))
        self.config.detrot = float(config.pop('detrot'))
        self.config.refractionindex = float(config.pop('refractionindex'))
        #self.config.air_absorbtion = float(config.pop('air_abs'))
        self.config.lossps = float(config.pop('lossps'))        
        
        #ub = config.pop('UB')
        crystalparam = np.array(util.parse_tuple(config.pop('crystal'), length=6, type=float))
        self.config.crystal = HKLVlieg.Crystal(crystalparam[:3],crystalparam[3:]) 
        self.config.L_range = np.array(util.parse_tuple(config.pop('l_range'), length=2, type=float))
        

        self.config.UB = HKLVlieg.UBCalculator(self.config.crystal,self.config.E_keV)

        #self.config.UB.setU(np.array(util.parse_tuple(config.pop('orient'), length=9, type=float)))
        
        self.config.dc = DetectorCalibration.DetectorCalibration(self.config.E_keV,self.config.crystal,self.config.px_size_mm*1e-3)
        self.config.dc.setCalibration([self.config.cen_x,self.config.cen_y],self.config.distance,self.config.detrot)
        
        self.config.cal_reflections = np.loadtxt(config.pop('cal_reflections'),comments='#').T
        if 'xmirror' in config:
            self.config.xmirror = util.parse_bool(config.pop('xmirror'))
        
        if 'th_offset' in config:
            self.config.th_offset = float(config.pop('th_offset'))
        else:
            self.config.th_offset = 0.0
        
        self.config.roiheight_start = float(config.pop('roiheight_start'))
        self.config.roiheight_min = float(config.pop('roiheight_min'))
        self.config.roiheight_rate = float(config.pop('roiheight_rate'))
        self.config.roiheightoffset = float(config.pop('roiheightoffset'))
        
        self.config.roiwidth = float(config.pop('roiwidth'))
        self.config.roiwidthoffset = float(config.pop('roiwidthoffset'))


    ####################################################################

    def get_destination_options(self, command):
        if not command:
            return False
        command = np.array(command,dtype=np.float)#util.parse_multi_range(','.join(command).replace(' ',','))
        scan = int(command[0])
        #rod = command[1:]
        #H = "%i_%02i" % (round(rod[0]), round((rod[0] % 1) * 100))
        #K = "%i_%02i" % (round(rod[1]), round((rod[1] % 1) * 100))
        
        return dict(scanno=scan)

    def getParam(self,specfile,scanno):
        sf = SpecFile(specfile)
        scan = sf[scanno-1]
        filename_temp = scan.scan_header_dict['C'].splitlines()[2]
        filename_base = filename_temp[18:-8]
        first_scan = int(filename_temp[-8:-4])
        scan = sf[scanno]
        ths = np.mean([scan.data_column_by_name('th_UpPos'),\
              scan.data_column_by_name('th_DownPos')],axis=0)
        filenames = [None]*len(ths)
        for i,filename in enumerate(filenames):
            filenames[i] = filename_base+str(first_scan+i).zfill(4)+'.cbf'
        return ths, filenames    
   
