# -*- coding: utf-8 -*-
"""
Created on Fri Apr 21 18:42:18 2017

@author: Timo
"""

import numpy as np
import numpy.linalg as LA
from scipy.linalg import svd
import scipy.optimize as opt
from scipy import special
from scipy.optimize._numdiff import approx_derivative
import numdifftools as nd
from math import *
import os
import copy
import matplotlib.pyplot as plt
from matplotlib import colors as colors
import configparser
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
    
def formatMeasurement_LatexFloat(x,dx,pad=1):
    
    roundedx, roundederr, prec = roundMeasurement(x,dx)
    
    valstr = "${:{pad}.{prec}f} \pm ".format(float(roundedx),prec=prec,pad=pad)
    
    errstr ="{:{pad}.{prec}f}$".format(float(roundederr),prec=prec,pad=pad)
    
    return valstr + errstr


def roundMeasurement(x,dx):
    firstdigit = 10**np.floor(np.log10(dx)) #mask first digit != 0

    smaller3 = np.array(dx < 3*firstdigit,dtype=np.float)
    
    errprec = 10**(np.floor(np.log10(dx)) - smaller3)
    
    roundederr = np.ceil(dx/errprec)*errprec
    
    roundedx = np.round(x/errprec)*errprec
    precision = int(-1*np.log10(errprec))
    
    return roundedx, roundederr, precision

def F_stat(chi1, chi2, p1, p2, n):
    nom = (chi1 - chi2)/(p2-p1)
    denom = chi2/(n - p2)
    return nom/denom
    
def leastsq_covariance(fun,x0):
    J = nd.Jacobian(fun)(x0)
    
    if J.ndim != 2:
        J = np.atleast_2d(J)
    _, s, VT = LA.svd(J, full_matrices=False)
    threshold = np.finfo(float).eps * max(J.shape) * s[0]
    s = s[s > threshold]
    VT = VT[:s.size]
    pcov = np.dot(VT.T / s**2, VT)
    return pcov

def leastsq_covariance_unstable(fun,x0):
    Jfun = nd.Jacobian(fun)
    J =  Jfun(x0)
    if J.ndim != 2:
        J = np.atleast_2d(J)
    pcov = LA.pinv(np.dot(J.T ,J))
    
    return pcov
    
def leastsq_covariance_lm(fun,x0):
    
    res = opt.least_squares(fun,x0,jac='3-point')
    J = res.jac
    if J.ndim != 2:
        J = np.atleast_2d(J)
    _, s, VT = LA.svd(J, full_matrices=False)
    threshold = np.finfo(float).eps * max(J.shape) * s[0]
    s = s[s > threshold]
    VT = VT[:s.size]
    pcov = np.dot(VT.T / s**2, VT)
    return pcov

	
def getScalefactor_fitted(ctr,xtal):
    h,k = ctr.hk
    harr = np.full_like(ctr.l,h)
    karr = np.full_like(ctr.l,k)
    logI = np.log( xtal.F(harr,karr,ctr.l)**2 )
    
    def minfun(x):
        if x[0] < 0:
            return np.inf
        return np.sum((np.log( (ctr.sfI*x[0])**2 ) - logI )**2)
    res = opt.minimize(minfun,[1.])
    return res.x

def getScalefactor(ctr,xtal):
    h,k = ctr.hk
    harr = np.full_like(ctr.l,h)
    karr = np.full_like(ctr.l,k)
    F_cryst = np.abs( xtal.F(harr,karr,ctr.l))
    return np.prod(F_cryst/ctr.sfI)**(1/ctr.l.size)
	
def getScalefactor_collection(coll,xtal):
	scale = []
	for ctr in coll:
		scale.append(getScalefactor(ctr,xtal))
	return scale
    
def stepup(z,A,sigma,mu):
    return A*0.5*(1 + special.erf(np.sqrt(0.5)*((z-mu)/(sigma))))

def stepdown(z,A,sigma,mu):
    return A*0.5*(1 - special.erf(np.sqrt(0.5)*((z-mu)/(sigma))))

# fpr Pt100
def averageCTRs_Pt100(CTRcoll, cutoff=2,nosymmetry_factor=3):
    #temporary fix!!
    #from datautils.xrayutils import CTRplotutil
    #newColl = CTRplotutil.CTRCollection()
    rodnames = []
    for ctr in CTRcoll:
        h, k = ctr.hk
        if abs(h) > abs(k):
            hk = abs(h), abs(k)
        else:
            hk = abs(k), abs(h)
        rodnames.append(hk)
        
    #CTRcoll_copy = copy.deepcopy(CTRcoll)
    rodnames = list(sorted(set(rodnames)))
    equivalent = []
    for n in rodnames:
        rodtpe = [c for c in CTRcoll if np.sum(np.array(n)**2) == np.sum(np.array(c.hk)**2) ]
        equivalent.append(rodtpe)

    return averageCTRs(equivalent, cutoff,nosymmetry_factor)
    
def averageCTRs(equivalent_list, cutoff=2,nosymmetry_factor=3):
    #temporary fix!!
    from datautils.xrayutils import CTRplotutil
    newColl = CTRplotutil.CTRCollection()
    for e in equivalent_list:
        e_curr = list(sorted(e,key=lambda x : x.sfI.size))
        #print(e_curr)
        l = e_curr[0].l
        err = e_curr[0].err**2
        sfI = e_curr[0].sfI
        count = np.ones_like(l)
        #print(e_curr)
        """
        first calculate the mean structure factor 
        and the resulting statistical error
        """
        for ctrno in range(1,len(e_curr)):
            #print(l)
            ctr = copy.deepcopy(e_curr[ctrno])
            val,i1,i2 = np.intersect1d(l,ctr.l,return_indices=True)
            mask1 = np.zeros_like(l, dtype=np.bool)
            mask1[i1] = True
            mask2 = np.zeros_like(ctr.l, dtype=np.bool)
            mask2[i2] = True
            err[mask1] += ctr.err[mask2]**2
            sfI[mask1] += ctr.sfI[mask2]
            count[mask1] += 1
            l = np.concatenate((l,ctr.l[~mask2]))
            err = np.concatenate((err,ctr.err[~mask2]**2))
            sfI = np.concatenate((sfI,ctr.sfI[~mask2]))
            count = np.concatenate((count,np.ones(np.sum(~mask2))))
            
            sortarg = np.argsort(l)
            
            l = l[sortarg]
            err = err[sortarg]
            sfI = sfI[sortarg]
            count = count[sortarg]
            
        sfI /= count # mean structure factor of symmetry equivalent rods
        err = np.sqrt(err/count) # statistical error
        """
        now calculate the variance of symmetry equivalent reflections
        """
        varerr = np.zeros_like(err)
        
        for ctrno in range(1,len(e_curr)):
            #print(l)
            ctr = copy.deepcopy(e_curr[ctrno])
            val,i1,i2 = np.intersect1d(l,ctr.l,return_indices=True)
            mask1 = np.zeros_like(l, dtype=np.bool)
            mask1[i1] = True
            mask2 = np.zeros_like(ctr.l, dtype=np.bool)
            mask2[i2] = True
            varerr[mask1] += (sfI[mask1] - ctr.sfI[mask2])**2
            
        varerr = np.sqrt(varerr/count) # systematic error
        no_symmetry_equivalent = varerr == 0.
        meanagreement = np.mean((varerr/sfI)[~no_symmetry_equivalent])
        varerr[no_symmetry_equivalent] = meanagreement*sfI[no_symmetry_equivalent]*nosymmetry_factor # no symmetry equivalent reflections available
        
        agreement_factor = varerr/sfI
        suffient_qualtity_indicator = sfI*cutoff > err
        
        final_error = np.sqrt(err**2 + varerr**2)
        
        hk = list(reversed(sorted([ctr.hk for ctr in e])))[0]
        newctr = CTRplotutil.CTR(hk,l,sfI,final_error)
        newctr.contributions = count
        newctr.systematicerr = varerr
        newctr.staterr = err
        newctr.quality_reflection = suffient_qualtity_indicator
        newctr.agreement_factor = agreement_factor
        newColl.append(newctr)
    return newColl

def openRodfile(name):
    with open(name,'r') as ctr:
        names = next(ctr).split('\t')[:-1]
        names[0] = names[0][2:]
        formats = [np.float for name in names]
        return np.loadtxt(ctr,dtype={'names': names,'formats': formats} )


def makeSurePathExists(path):
    if not os.path.exists(path):
        os.makedirs(path)

def as_ndarray(obj):
    """make sure a float, int, list of floats or ints,
    or tuple of floats or ints, acts as a numpy array
    """
    if isinstance(obj, (float, int)):
        return np.array([obj])
    return np.asarray(obj)
    
def readNDarrayConfig(filename):
    
    header = StringIO()
    
    with open(filename,'r') as f:
        for line in f:
            if not line.startswith('#'):
                break
            header.write(line[2:])
    config = configparser.ConfigParser()
    config.read_string(header.getvalue())
    
    data = np.loadtxt(filename)
    return data.T, config
    
    

def sumRoi(image,roi):
    xlo,xhi = roi[0]
    ylo,yhi = roi[1]
    image_tmp = image[xlo:xhi,ylo:yhi]
    I = np.nansum(image_tmp)
    pixel = np.count_nonzero(~np.isnan(image_tmp))
    return I, pixel

def calcIntersect(g1,g2):
    g1P1 , g1P2 = g1
    g2P1 , g2P2 = g2
    
    a = g1P1 - g1P2
    b = g2P1 - g2P2
    
    res = LA.solve(np.array([a,-b]),g2P1-g1P1)
    
    return a*res[0] + g1P1
"""
gives delta in:
    
"""

def solveTrigEquation(theta1,theta2,l1,l2):
    
    def Chi2(delta):
        res = function(theta1,theta2,l1,l2,delta)
        #print("%s : %s" % (delta,res))
        return res
    
    res = opt.minimize(Chi2,np.array([0.0]),bounds=[(-np.pi,np.pi)],method='TNC')
    print(res)
    print(np.rad2deg(res.x))
    return res.x

def function(theta1,theta2,l1,l2,delta):
    t1 = cos(theta1 - delta[0])**2 * l1**2 
    t2 = cos(theta1 + delta[0])**2 * l2**2
    t3 = 2 * l1 * l2 * cos(theta1 + delta[0]) * cos(theta1 - delta[0]) * cos(2*theta2)
    t4 = ( (l2**2) * cos(theta1 + delta[0])**2 * sin(2*theta2)**2 )/( cos(delta[0] - theta2)**2 )
    return fabs(t1 + t2 - t3 - t4)

#solveTrigEquation(np.deg2rad(2.),np.deg2rad(76.),1.,1.1)

"""
only single value of th:
"""
def x_rotation(th):
    return np.matrix(((1., 0., 0.), (0., np.cos(th), -np.sin(th)), (0., np.sin(th), np.cos(th))))


def y_rotation(th):
    return np.matrix(((np.cos(th), 0., np.sin(th)), (0, 1., 0.), (-np.sin(th), 0., np.cos(th))))


def z_rotation(th):
    return np.matrix(((np.cos(th), -np.sin(th), 0.), (np.sin(th), np.cos(th), 0.), (0., 0., 1.)))

"""
th is now a 1d-array, returns an array of rotation matrices
"""
def x_rotationArray(th):
    matrices = np.zeros((th.size,3,3))
    matrices[:,0,0] = 1
    matrices[:,1,1] = np.cos(th)
    matrices[:,1,2] = -np.sin(th)
    matrices[:,2,1] = np.sin(th)
    matrices[:,2,2] = np.cos(th)
    return matrices
    

def y_rotationArray(th):
    matrices = np.zeros((th.size,3,3))
    matrices[:,1,1] = 1
    matrices[:,0,0] = np.cos(th)
    matrices[:,0,2] = np.sin(th)
    matrices[:,2,0] = -np.sin(th)
    matrices[:,2,2] = np.cos(th)
    return matrices


def z_rotationArray(th):
    matrices = np.zeros((th.size,3,3))
    matrices[:,2,2] = 1
    matrices[:,0,0] = np.cos(th)
    matrices[:,0,1] = -np.sin(th)
    matrices[:,1,0] = np.sin(th)
    matrices[:,1,1] = np.cos(th)
    return matrices

def orthogonal(matrix):
    matrix = np.matrix(matrix)
    
    SMALL = 1e-4
    
    def normalise(m):
        d = LA.norm(m)
        if d < SMALL:
            raise Exception("Error: can't make matrix orthogonal")
        return m / d
    #print(LA.norm(v1))
    v1 = normalise(matrix[:,0])
    print(matrix[:,0])
    print(LA.norm(v1))
    v2 = normalise(matrix[:,1])
    v3 = normalise(matrix[:,2])
    
    return np.hstack([v1, v2, v3]).A


def calcHighPixel(image,threshold):
    return np.argwhere(image > threshold)
    
    #old:
    highintensity = image > threshold
    highpixel = []

    for x,xrow in enumerate(highintensity):
        for y in range(xrow.size):
            if highintensity[x][y]:
                highpixel.append([y,x])

    highpixel = np.array(highpixel).T
    return highpixel

def plotP3Image(image,vmin=0,vmax=None,cmap='jet',thresholdMarker=None,**keyargs):
    if 'figure' in keyargs:
        fig = keyargs['figure']
        ax = fig.add_subplot(111)
    elif 'axis' in keyargs:
        ax = keyargs['axis']
        fig = ax.get_figure()
    else:
        fig = plt.figure(figsize=(12,14))
        ax = fig.add_subplot(111)

    if not vmax:
        ax.imshow(image,interpolation='none',cmap=plt.get_cmap(cmap),norm=colors.SymLogNorm(linthresh=1,linscale=1,vmin=vmin))
    else:
        ax.imshow(image,interpolation='none',cmap=plt.get_cmap(cmap),norm=colors.SymLogNorm(linthresh=1,linscale=1,vmin=vmin,vmax=vmax))
    
    if thresholdMarker is not None:
        highpixel = calcHighPixel(thresholdMarker)
        if highpixel.size > 0:
            ax.plot(highpixel[0],highpixel[1],'ro')
    ax.set_ylim([1700,0])
    ax.set_xlim([0,1475])
    numrows, numcols = image.shape
    def format_coord(x, y):
        col = int(x + 0.5)
        row = int(y + 0.5)
        if col >= 0 and col < numcols and row >= 0 and row < numrows:
            z = image[row, col]
            return 'x=%1.4f, y=%1.4f, z=%1.4f' % (x, y, z)
        else:
            return 'x=%1.4f, y=%1.4f' % (x, y)
    ax.format_coord = format_coord
    return fig, ax

#def deltaGamma(drr,gamma):
#    return np.arctan((1 + drr) / np.abs(np.cos(gamma))) - gamma