# -*- coding: utf-8 -*-
from .HKLVlieg import Crystal, UBCalculator, VliegAngles, GeometryCorrection
import numpy as np
import numpy.linalg as LA
from .. import util
import xraydb
import warnings
import random
import json
import math
from scipy.interpolate import interp1d
from scipy.signal import savgol_filter
from scipy.ndimage import gaussian_filter1d
import os
import re
from scipy.special import erf
random.seed(45)
from collections import OrderedDict

filepath = os.path.split(__file__)[0]


class SXRDCrystal(object):
    def __init__(self,uc_bulk,*uc_surface,**keyargs):
        self.uc_bulk = uc_bulk
        self.uc_surface_list = list(uc_surface)
        self.domains = [[(np.identity(3),1)] for i in self.uc_surface_list]
        if 'atten' in keyargs:
            self.atten = keyargs['atten']
        else:
            self.atten = 0.01 
        self.weightparameters = []
        self.weightparameterLimits = []
        self.weights = np.ones(len(self.uc_surface_list))/len(self.uc_surface_list)
        self.werrors = None
        self.fitparameters = [[] for i in self.uc_surface_list]
        self.fitp_limits = []
        self.fitp_values = np.array([])
        self.numberPars = 0
        
        
    def setGlobalReferenceUnitCell(self,uc_reference,rotMatrix=np.identity(3)):
        """   
        sets a global reference unit cell. 
        When any F_hkl is called, first hkl will be transformed from the reference 
        lattice into the lattice of this unit cell,
        using equation
        B' * H' = O * B * H, where a ' indicates the entities in the frame of this
        lattice. O is an optional rotation matrix, which describes the rotation of
        this lattice with respect to the reference lattice.
        
        Equally the same applies to the relative coordinates of the respective 
        lattices:
        R' * x' = O * R * x
        
        """
        self.uc_bulk.setReferenceUnitCell(uc_reference,rotMatrix)
        [uc.setReferenceUnitCell(uc_reference,rotMatrix) for uc in self.uc_surface_list]
    
    def F_surf(self,harray,karray,Larray):
        F = np.empty_like(Larray,dtype=np.complex128)
        for uc, weight in zip(self.uc_surface_list,self.weights):
            F += weight*uc.F_uc(harray,karray,Larray)
        return F
    
    def F(self,harray,karray,Larray):
        F = self.uc_bulk.F_bulk(harray,karray,Larray,self.atten)
        hkl = np.vstack((harray,karray,Larray))
        for uc, weight, domains in zip(self.uc_surface_list,self.weights,self.domains):
            for matrix, occup in domains:
                hkl_n = np.dot(matrix,hkl)
                F += occup*weight*uc.F_uc(hkl_n[0],hkl_n[1],hkl_n[2])
        return F
    

    def setDomain(self,uc_no,domains):
        """
        should be a list of tuples with:
        (rotmatrix, occupancy)
        """
        self.domains[uc_no] = domains
    
    def addRelParameter(self,index0_or_name,indexneg_or_name=None,initial=0.5,limits=[-np.inf,np.inf]):
        """
        sets a new fit parameter for the weight w of a unit cell
        
        F_xtal = w * F_uc1
        
        if indexneg_or_name is not None the structure factor will be calculated
        to
        
        F_xtal = w * F_uc1 +  (1 - w) * F_uc2
        
        Parameters
        ----------
        index0_or_name : str or int
            index or name of the unit cell 1 (with F_uc1)
        indexneg_or_name : default: None, str or int
            index or name of the unit cell 2 (with F_uc2)
        initial: float
            initial value of w
        limits: list
            fit limits of w 
        """
        index0 = self.getUcIndex(index0_or_name)
        if indexneg_or_name is not None:
            indexneg = self.getUcIndex(indexneg_or_name)
        else:
            indexneg = None
        self.weightparameters.append([index0,indexneg,initial])
        self.weightparameterLimits.append(limits)
        self.applyParameters()
        
    #def addUCWeightParameter(self,index_or_name,limits=[-np.inf,np.inf]):
        
        
        
    def applyParameters(self):
        for [index0,indexneg,value] in self.weightparameters:
            self.weights[index0] = value
            if indexneg is not None:
                self.weights[indexneg] = 1. - value
    
    def getInitialParameters(self):
        x0 = []
        bp = self.uc_bulk.getInitialParameters()
        self.bulk_param = bp.size
        x0.append(bp)
        
        self.uc_param = []
        for i, uc in enumerate(self.uc_surface_list):
            uc_p = list(uc.getInitialParameters())
            for par in reversed(self.fitparameters[i]):
                self.fitparameters[i][2] = uc_p[int(par[1])]
                del uc_p[int(par[1])]
            x0.append(np.array(uc_p))
            self.uc_param.append(len(uc_p))
        if self.numberPars > 0:
            self._readFitparValues()
            x0.insert(0,self.fitp_values)
            
        x_rel = np.empty(len(self.weightparameters),dtype=np.float)
        for i,[index0,indexneg,value] in enumerate(self.weightparameters):
            x_rel[i] = value  
        x0.append(x_rel)

        return np.concatenate(x0)
    
    def getStartParamAndLimits(self):
        x0 = []
        lower = []
        higher = []
        bp, low, high = self.uc_bulk.getStartParamAndLimits()
        self.bulk_param = bp.size
        x0.append(bp)
        lower.append(low)
        higher.append(high)
        
        self.uc_param = []
        for i, uc in enumerate(self.uc_surface_list):
            uc_p, low, high = [list(x) for x in uc.getStartParamAndLimits()]
            for par in reversed(self.fitparameters[i]):
                par[2] = uc_p[int(par[1])]
                del uc_p[int(par[1])]
                del low[int(par[1])]
                del high[int(par[1])]
            self.uc_param.append(len(uc_p))
            x0.append(np.array(uc_p))
            lower.append(np.array(low))
            higher.append(np.array(high))
        if self.numberPars > 0:
            self._readFitparValues()
            low, upp = np.array(self.fitp_limits).T
            x0.insert(0,self.fitp_values)
            lower.insert(0,low)
            higher.insert(0,upp)
        x_rel = np.empty(len(self.weightparameters),dtype=np.float)
        lo_rel = np.empty(len(self.weightparameters),dtype=np.float)
        hi_rel = np.empty(len(self.weightparameters),dtype=np.float)
        for i,[index0,indexneg,value] in enumerate(self.weightparameters):
            x_rel[i] = value  
            lo_rel[i] = self.weightparameterLimits[i][0]
            hi_rel[i] = self.weightparameterLimits[i][1]
        x0.append(x_rel)
        lower.append(lo_rel)
        higher.append(hi_rel)
        return np.concatenate(x0), np.concatenate(lower), np.concatenate(higher)
    
    def _readFitparValues(self):
        """
        updates fitp_values from fitparameters
        
        reads parameter values in self.fitparameters and writes 
        the mean value to the corresponding entry in fitp_values

        Returns
        -------
        None.

        """
        values = np.zeros(len(self.fitp_values))
        count = np.zeros(len(self.fitp_values))
        for uc_pars in self.fitparameters:
            for p in uc_pars:
                values[int(p[0])] += p[2]
                count[int(p[0])] += 1.
            
        self.fitp_values = values/count
        

    def addFitParameter(self,parameter,limits=[-np.inf,np.inf]):
        newParIndex = self.numberPars
        for uc_identifier in parameter:
            indexarray = parameter[uc_identifier]
            uc_index = self.getUcIndex(uc_identifier)
            parameterindex_uc = self.uc_surface_list[uc_index].addFitParameter(indexarray,limits)
            self.fitparameters[uc_index].append(np.array([newParIndex,parameterindex_uc,np.nan]))
        self.numberPars += 1
        self.fitp_limits.append(limits)
        self.fitp_values = np.append(self.fitp_values,np.nan)
        
        for par in self.fitparameters:
            par.sort(key=lambda x : x[1])
        
        return self.numberPars - 1
            
    
    def setParameters(self,x):
        self.fitp_values = x[:self.numberPars]
        lower = self.numberPars
        self.uc_bulk.setFitParameters(x[lower:self.bulk_param+lower])
        lower = self.bulk_param + lower
        for uc,noparam,params in zip(self.uc_surface_list,self.uc_param,self.fitparameters):
            x_uc = list(x[lower:noparam + lower])
            for p in params:
                x_uc.insert(int(p[1]),self.fitp_values[int(p[0])])
            uc.setFitParameters(x_uc)
            lower = noparam + lower
        remainder = x[lower:]
        for i,par in enumerate(self.weightparameters):
            par[2] = remainder[i]
        self.applyParameters()
        
    def setFitErrors(self,errors):
        self.fitp_errors = errors[:self.numberPars]
        lower = self.numberPars
        self.uc_bulk.setFitErrors(errors[lower:self.bulk_param+lower])
        lower = self.bulk_param + lower
        for uc,noparam,params in zip(self.uc_surface_list,self.uc_param,self.fitparameters):
            errors_uc = list(errors[lower:noparam + lower])
            for p in params:
                errors_uc.insert(int(p[1]),self.fitp_errors[int(p[0])])
            uc.setFitErrors(errors_uc)
            lower = noparam + lower
        remainder = errors[lower:]
        self.werrors = np.full_like(self.weights,np.nan)
        for i,[index0,indexneg,value] in enumerate(self.weightparameters):
            self.werrors[index0] = remainder[i]
            self.werrors[indexneg] = remainder[i]
       
    def setEnergy(self,E):
        for uc in self.uc_surface_list:
            uc.setEnergy(E)
        self.uc_bulk.setEnergy(E)
        
        
    def zDensity_G(self,z,h,k):
        rho = self.uc_bulk.zDensity_G_asbulk(z,h,k)
        for uc,w in zip(self.uc_surface_list,self.weights):
            rho += uc.zDensity_G(z,h,k)*w
        return rho
        
        
    def toRODStr(self):
        s = "E = %.5f keV\n" % (self.uc_bulk._E*1e-3)
        for i,(w,uc) in enumerate(zip(self.weights,self.uc_surface_list)):
            s += "# %s %s\n" % (uc.__class__.__name__ ,uc.name)
            s += "%04i %.5f\n" % (i,w)
            s += uc.toRODStr() + "\n"
            
        s+= "# UnitCell bulk\n" + self.uc_bulk.toRODStr()
        return s
    
    def toStr(self,showErrors=True):
        s = "E = %.5f keV\n" % (self.uc_bulk._E*1e-3)
        for i,(w,uc) in enumerate(zip(self.weights,self.uc_surface_list)):
            s += "# %s %s\n" % (uc.__class__.__name__ ,uc.name)
            if showErrors and self.werrors is not None:
                s += "%04i occupancy = %.5f +- %.5f\n" % (i,w,self.werrors[i])
            else:
                s += "%04i occupancy = %.5f\n" % (i,w)
            s += uc.toStr() + "\n"
            
        s+= "# UnitCell bulk\n" + self.uc_bulk.toStr()
        return s
    

    def fromStr(string,atten=0.01):
        weights = []
        werrors = []
        uc_suface = []
        
        strio = util.StringIO(string)
        Estr = next(strio)
        E = float(Estr[4:-4])*1e3
        string = strio.read()
        errors = False
        unitCells = string.split('#')[1:]
        for ucstr in unitCells[:-1]:
            strio = util.StringIO(ucstr)
            line1 = next(strio)
            classname,name = line1.split(maxsplit=1)
            name = name.strip()
            line2 = next(strio).split()
            if '+-' in line2:
                errors = True
                weights.append(float(line2[line2.index('+-')-1]))
                werrors.append(float(line2[line2.index('+-')+1]))
            else:
                weights.append(float(line2[1]))
                werrors.append(float('nan'))
            if classname == 'UnitCell':
                uc = UnitCell.fromStr(strio.read())
            elif classname == 'WaterModel':
                uc = WaterModel.fromStr(strio.read())
            else:
                raise NotImplementedError("class name not understood: %s" % classname)
            uc.name = name
            uc.setEnergy(E)
            uc_suface.append(uc)
        bulkstrio = util.StringIO(unitCells[-1])
        next(bulkstrio)
        uc_bulk = UnitCell.fromStr(bulkstrio.read())
        uc_bulk.setEnergy(E)
        xtal = SXRDCrystal(uc_bulk,*uc_suface,atten=atten)
        xtal.weights = np.array(weights)
        xtal.werrors = np.array(werrors) if errors else None
        return xtal
        
    def __getitem__(self,uc_name_or_index):
        if isinstance(uc_name_or_index,int):
            return self.uc_surface_list[uc_name_or_index]
        elif isinstance(uc_name_or_index,str):
            if uc_name_or_index == 'bulk':
                return self.uc_bulk
            namelist = [uc.name for uc in self.uc_surface_list]
            idx = namelist.index(uc_name_or_index)
            return self.uc_surface_list[idx]
        else:
            raise ValueError("must be str or int, not {}".format(type(uc_name_or_index)) )
            
    def getUcIndex(self,uc_name_or_index):
        if isinstance(uc_name_or_index,int):
            return uc_name_or_index
        elif isinstance(uc_name_or_index,str):
            namelist = [uc.name for uc in self.uc_surface_list]
            idx = namelist.index(uc_name_or_index)
            return idx
        else:
            raise ValueError("must be str or int, not {}".format(type(uc_name_or_index)) )        
            
            
    def getUcNames(self):
        return [uc.name for uc in self.uc_surface_list] + ['bulk']

    def toFile(self,filename):
        with open(filename,'w') as f:
            f.write(self.toRODStr())
        
    def fromFile(filename):
        with open(filename,'r') as f:
            fstr = f.read()
        return SXRDCrystal.fromStr(fstr)
    
    def __repr__(self):
        return self.toStr()

class WaterModel(Crystal):
    
    supportedTypes = ['1layer_O','layered_O','1layer_s_O']
    
    parameterOrder = "zpos/frac layer_spacing/frac sigma_0/frac  sigma_bar/frac rho_0_r"
    
    parameterLookup = {'zpos/frac' : 0, 'layer_spacing/Angstrom' : 1,
                       'sigma_0/frac' : 2, 'sigma_bar/frac' : 3, 'rho_0_r' : 4}
    
    def __init__(self,a,alpha,watertype='step',**keyargs):
        super(WaterModel,self).__init__(a,alpha)
        self.basis = [0.,2.8,0.5,1.,0.]
        self.pw = 0.0334
        self.fitparameters = []
        self.relfitparam = []
        self.fitparlimits = []
        self.relfitlimits = []
        self.basis_0 = None
        self.errors = None
        if 'name' in keyargs:
            self.name = keyargs['name']
        else:
            self.name = "unnamed water model"
        if watertype not in WaterModel.supportedTypes:
            raise Exception("Water model %s is not supported, should be one of %s" % (watertype,WaterModel.supportedTypes))
        self.type = watertype
            
        self.refHKLTransform = np.matrix(np.identity(3))
        self.refRealTransform = np.matrix(np.identity(3))
        a,alpha,_,_ = self.getLatticeParameters()
        self.uc_area = a[0]*a[1]*np.sin(alpha[2]) #unit cell area
    
    def setReferenceUnitCell(self,uc,rotMatrix=np.identity(3)):
        """   
        set reference unit cell. 
        When any F_hkl is called, first hkl will be transformed from the reference 
        lattice into the lattice of this unit cell,
        using equation
        B' * H' = O * B * H, where a ' indicates the entities in the frame of this
        lattice. O is an optional rotation matrix, which describes the rotation of
        this lattice with respect to the reference lattice.
        
        Equally the same applies to the relative coordinates of the respective 
        lattices:
        R' * x' = O * R * x
        
        !!! This has to be checked !!!
        """
        self.refRealTransform = self.RealspaceMatrix.I * rotMatrix * uc.RealspaceMatrix 
        self.refHKLTransform = self.getB().I * rotMatrix * uc.getB()

    def F_uc(self,h,k,l):
        mask = np.logical_and(np.isclose(h,0), np.isclose(k,0))
        if not np.any(mask):
            return np.zeros_like(l,dtype=np.complex128)
        a,alpha,b,beta = self.getLatticeParameters()
        l_masked = l[mask]
        zpos, d_layering, sigma_0, sigma_bar, A0 = self.basis
        
        
        if self.type == 'layered_O':
            
            f = np.empty_like(l_masked,dtype=np.complex128)
            f[:] =  self.f[0][10] + self.f[0][11] + 1j*self.f[0][12]
            for j in range(5):
                f += self.f[0][j]*np.exp(- self.f[0][j+5]*(l_masked*b[2])**2)
            
            f_w_layer = f*self.pw*self.volume*d_layering
            
            if A0 == 1.:
                sf = np.exp(-2*(np.pi * sigma_0 * l_masked)**2)/(1. - np.exp(-2*(np.pi * sigma_bar * l_masked)**2)*np.exp(2j*np.pi * l_masked * d_layering))
                F_wat = f_w_layer*sf
            else:
                sigma_1 = np.sqrt(sigma_0**2 + sigma_bar**2)
                sf = np.exp(-2*(np.pi * sigma_1 * l_masked)**2)/(1. - np.exp(-2*(np.pi * sigma_bar * l_masked)**2)*np.exp(2j*np.pi * l_masked * d_layering))
                F_wat = f_w_layer*( A0*np.exp(-0.5*(sigma_0*l_masked*b[2])**2) + sf*np.exp(2j*np.pi * l_masked * d_layering))
            
        elif self.type == '1layer_O':
            rho, mu_offset = self._1layer_firstGauss(0.00)
            f = np.empty_like(l_masked,dtype=np.complex128)
            f[:] =  self.f[0][10] + self.f[0][11] + 1j*self.f[0][12]
            for j in range(5):
                f += self.f[0][j]*np.exp(- self.f[0][j+5]*(l_masked*b[2])**2)
                
            f *= (10/8)
            F_wat = 1j*self.pw*f*self.volume*np.exp(-2*(np.pi * sigma_0 * l_masked)**2)/(2*np.pi*l_masked)
                            
            F_wat = F_wat + A0*f*np.exp(-0.5*(sigma_0*l_masked*b[2])**2)*np.exp(2j*np.pi * l_masked * mu_offset)
            
        elif self.type == '1layer_s_O':
            rho, mu_offset = self._1layer_firstGauss(0.0)
            f = np.empty_like(l_masked,dtype=np.complex128)
            f[:] =  self.f[2][10] + self.f[2][11] + 1j*self.f[2][12] # self.f[2] is O2-
            for j in range(5):
                f += self.f[2][j]*np.exp(- self.f[2][j+5]*(l_masked*b[2])**2)
                
            F_wat = 1j*self.pw*f*self.volume*np.exp(-2*(np.pi * sigma_0 * l_masked)**2)/(2*np.pi*l_masked)
                            
            F_wat = F_wat + A0*f*np.exp(-0.5*(sigma_0*l_masked*b[2])**2)
            
        else:
            warnings.warn("unknown water type %s !" % self.type)
            F_wat = np.zeros_like(l_masked)

        F_wat *= np.exp(2j*np.pi * l_masked * zpos) # position of water
        
        #temp = np.exp(-0.5*(Q*sigma_bar)**2)*np.exp(1j*Q*spacing) 
        #flayer = spacing*(zwater)*pw*self.uc_area*np.exp(-0.5*(Q*sigma_0)**2)
        #F = (np.exp(2j*np.pi*l[mask]*zpos)*(flayer/(1-temp)))/self.volume
        
        F_water = np.zeros_like(l,dtype=np.complex128)
        F_water[mask] = F_wat
        return F_water/self.volume
    
    def setWaterParameters(self,zpos,layer_spacing,sigma_0,sigma_bar,A0):
        self.basis = [zpos,layer_spacing,sigma_0,sigma_bar,A0]

    def setEnergy(self,E):
        self._E = E
        self.lookupScatteringFactors(E)

    def lookupScatteringFactors(self,E):        
        self.f = np.empty((3,13),dtype=np.float64)
        for i,name in enumerate(['O','H','O2-']):
            self.f[i,:11] = readWaasmaier(name)
            self.f[i,11:] = readDispersion(name,E)
        f1,f2 = UnitCell.special_formfactors['H2O'][1](E)
        self.wat_dispersion = f1 + 1j*f2

        

    def addFitParameter(self,indexarray,limits=[-np.inf,np.inf]):
        if not limits[0] <= np.mean(self.basis[indexarray]) <= limits[1]:
            raise ValueError("start parameter is not within limits!")  
        self.fitparameters.append(indexarray)
        self.fitparlimits.append(limits)
        return len(self.fitparameters) - 1


    def addRelParameter(self,indexarray,factors,limits=[-np.inf,np.inf]):
        #if not limits[0] <= np.mean(self.basis[indexarray]) <= limits[1]:
        #    raise ValueError("start parameter is not within limits!")  
        self.relfitparam.append((indexarray,factors))
        self.relfitlimits.append(limits)
        return len(self.relfitparam) - 1

    def fixInitialBasis(self):
        self.basis_0 = np.copy(self.basis)

    def getInitialParameters(self):
        if self.basis_0 is None:
            self.basis_0 = np.copy(self.basis)
        
        x0 = np.empty(len(self.fitparameters) + len(self.relfitparam),dtype=np.float)
        
        for i,parray in enumerate(self.fitparameters):
            x0[i] = np.mean(self.basis[parray])
        
        for i,(parray,factor) in enumerate(self.relfitparam,len(self.fitparameters)):
            x0[i] = np.mean( (self.basis[parray] - self.basis_0[parray]) / factor)
        return x0         
        
    def getStartParamAndLimits(self):
        x0 = self.getInitialParameters()
        if x0.size > 0:
            lower, upper = np.array(self.fitparlimits + self.relfitlimits).T
        else:
            lower, upper = np.array([]),  np.array([])
        return x0, lower, upper
    
    def setFitParameters(self,x):
        x_0 = x[:len(self.fitparameters)]
        x_r = x[len(self.fitparameters):]
        for val, parray in zip(x_0,self.fitparameters):
            self.basis[parray] = val
        for val, (parray,factor) in zip(x_r,self.relfitparam):
            self.basis[parray] = factor*val + self.basis_0[parray]
            
    def _1layer_firstGauss(self,biaswidth=0.2547):
        p = [ 0.85886939,  1.03841354, -1.35438352]
        zpos, d_layering, sigma_0, sigma_bar, A0 = self.basis
        #a,alpha,_,_ = self.getLatticeParameters()
        #uc_area = a[0]*a[1]*np.sin(alpha[2])
        
        rho_0 = A0/(self.volume*np.sqrt(2*np.pi*(sigma_0**2 + (biaswidth/self._a[2])**2)))
        
        A = rho_0/self.pw
        
        mu_rel = np.sqrt(2*np.log(2)) + util.stepdown(np.log10(A),p[0],p[1],p[2])
        mu = mu_rel*np.sqrt(sigma_0**2 + (biaswidth/self._a[2])**2)
        
        return rho_0*(10.+self.wat_dispersion), mu
    
        
        
        
    def zDensity_G(self, z, h,k):
        """
        calculates h,k-th Fourier component of the electron density
        Water models only contribute to the 0,0-th component
        Will return an array with zeros for h, k != 0 
        
        The density is normalized to the surface area of the unit cell

        Parameters
        ----------
        z : 1-d array
            z coordinates in Angstrom, should be equidestant 
            and monotonally increasing to avoid numerical issues with convolutions
        h : float
            h-th component index
        k : float
            k-th component index

        Returns
        -------
        1d- array complex128
            complex h,k-th Fourier component of the electron density 
            in electrons/Angstrom**3

        """    
        
        if abs(h) > 0.001 or  abs(k) > 0.001:
            return np.zeros_like(z)
        zstep = np.diff(z)
        zstep_mean = np.mean(zstep)
        if not np.allclose(zstep,zstep_mean):
            warnings.warn("zDensity: z stepsize is not equal in given z array."
                          "This will result in numerical errors in electron density calculation!")
        zpos, d_layering, sigma_0, sigma_bar, A0 = self.basis
        if self.type == 'layered_O':
            zpos, d_layering, sigma_0, sigma_bar, A0 = self.basis
            zrange_waterstructure = np.amax(z)/self._a[2] - zpos # lattice units
            nogausseans_inrange = np.ceil(zrange_waterstructure/d_layering)
            nogausseans = int(nogausseans_inrange + 5) # some additional gausseans to reduce edge effects
            layer_density = self.pw*d_layering*self._a[2]
            rho = np.zeros_like(z,dtype=np.complex128)
            
            for i in range(nogausseans):
                sigma2_rel = sigma_0**2 + i*sigma_bar**2
                deltaZ2i = sigma2_rel*self._a[2]**2
                
                z_i = (zpos + i*d_layering)*self._a[2]
                
                rho_i = ((self.f[0][10] + self.f[0][11] + 1j*self.f[0][12]) / (np.sqrt(2*np.pi*deltaZ2i))) *np.exp( -0.5 *(  ((z- z_i)**2)/deltaZ2i ))
                for j in range(5):
                    exp_dz = self.f[0][j+5] + 0.5*deltaZ2i
                    rho_i += (self.f[0][j]/(np.sqrt(4.*np.pi*exp_dz))) * np.exp(- (((z- z_i)**2)/(4*exp_dz))  )
                #rho_i *= np.exp(-2j*np.pi*((z- z_i)/self._a[2]))
                if i == 0:
                    rho += A0*rho_i*layer_density
                else:
                    rho += rho_i*layer_density
        elif self.type == '1layer_O':

            

            #rho = 0.5*(10.+self.wat_dispersion)*self.pw*(1. + erf((z - zpos*self._a[2])/(np.sqrt(2)* sigma_0 * self._a[2])))
            #rho = gaussian_filter1d(np.abs(rho),0.2547 / zstep_mean).astype(np.complex128) #estimation of water molecular form factor
            rho_layer, mu_offset = self._1layer_firstGauss(0.00)
            
            deltaZ2i = (sigma_0*self._a[2])**2
            z_i = (zpos + mu_offset)*self._a[2]
            
            rho_i = ((self.f[0][10] + self.f[0][11] + 1j*self.f[0][12]) / (np.sqrt(2*np.pi*deltaZ2i))) *np.exp( -0.5 *(  ((z- z_i)**2)/deltaZ2i ))
            rho_step = util.stepup(z,self.f[0][10] + self.f[0][11] + 1j*self.f[0][12],sigma_0*self._a[2],zpos*self._a[2])
            for j in range(5):
                #exp_dpara = self.f[i][j+5] + 0.5*deltaPara2i
                exp_dz = self.f[0][j+5] + 0.5*deltaZ2i
                eff_sigma = np.sqrt(exp_dz/2)
                rho_step += util.stepup(z,self.f[0][j],eff_sigma,zpos*self._a[2])
                rho_i += (self.f[0][j]/(np.sqrt(4.*np.pi*exp_dz))) * np.exp(- (((z- z_i)**2)/(4*exp_dz))  )
            rho_i *= A0*np.exp(-2j*np.pi*((z- z_i)/self._a[2]))
            
            rho_step *= (10/8)*self.pw
            
            rho = rho_i/self.uc_area + rho_step
        elif self.type == '1layer_s_O':
            rho_layer, mu_offset = self._1layer_firstGauss(0.0)
            
            deltaZ2i = (sigma_0*self._a[2])**2
            z_i = (zpos)*self._a[2]
            
            rho_i = ((self.f[2][10] + self.f[2][11] + 1j*self.f[2][12]) / (np.sqrt(2*np.pi*deltaZ2i))) *np.exp( -0.5 *(  ((z- z_i)**2)/deltaZ2i ))
            rho_step = util.stepup(z,self.f[2][10] + self.f[2][11] + 1j*self.f[2][12],sigma_0*self._a[2],zpos*self._a[2])
            for j in range(5):
                #exp_dpara = self.f[i][j+5] + 0.5*deltaPara2i
                exp_dz = self.f[2][j+5] + 0.5*deltaZ2i
                eff_sigma = np.sqrt(exp_dz/2)
                rho_step += util.stepup(z,self.f[2][j],eff_sigma,zpos*self._a[2])
                rho_i += (self.f[2][j]/(np.sqrt(4.*np.pi*exp_dz))) * np.exp(- (((z- z_i)**2)/(4*exp_dz))  )
            rho_i *= A0*np.exp(-2j*np.pi*((z- z_i)/self._a[2]))
            
            rho_step *= self.pw
            
            rho = rho_i/self.uc_area + rho_step
            
            
        return rho


    def setFitErrors(self,errors):
        self.errors = np.full_like(self.basis,np.nan)
        err_0 = errors[:len(self.fitparameters)]
        err_r = errors[len(self.fitparameters):]
        for val, parray in zip(err_0,self.fitparameters):
            self.errors[parray] = val
        for val, (parray,factor) in zip(err_r,self.relfitparam):
            self.errors[parray] = factor*val

    def __repr__(self):
        repr_super = super(WaterModel,self).__repr__()
        return "%s\n water model" % repr_super
        
    def waterModelToStr(self,showErrors=True):
        param = self.basis
        if (self.errors is not None) and showErrors:
            err = self.errors
            l = []
            for t in zip(param,err):
                [l.append(ti) for ti in t]
            return "Water layer  (%.5f +- %.5f)  (%.5f +- %.5f)  (%.5f +- %.5f)  (%.4f +- %.4f)  (%.4f +- %.4f)" % tuple(l)
        else:
            return "%.5f     %.5f     %.5f  %.5f  %.5f" % tuple(param)

    def latticeRODStr(self):
        a,alpha,_,_ = self.getLatticeParameters()
        return "return\n%s\n%.4f %.4f %.4f %.4f %.4f %.4f" % (self.type,*a,*np.rad2deg(alpha))
    
    
    def parameterStr(self):
        return self.waterModelToStr() + "\n"
    
    def parameterStrRod(self):
        return self.waterModelToStr(False) + "\n"
            
    def __str__(self):
        st = repr(self) + "\n"
        st += "id  " + WaterModel.parameterOrder + "\n"
        return st + self.parameterStr()
            
    def writeWATfile(self,filename):
        with open(filename,'w') as f:
            st = self.latticeRODStr() + "\n" + self.parameterStrRod()
            f.write(st)
            
    def toRODStr(self):
        return self.latticeRODStr() + "\n" + self.parameterStrRod()
    
    def toStr(self):
        return self.latticeRODStr() + "\n" + WaterModel.parameterOrder + "\n" + self.parameterStr()
    
    @staticmethod
    def fromWATfile(watfile):
        with open(watfile,'r') as f:
            next(f) # return line
            nline = next(f).split()
            try:
                latticeparams = np.array(nline,dtype=np.float64)
                watertype = 'step' # default behaviour
            except ValueError:
                watertype = nline[0]
                latticeparams = np.array(next(f).split(),dtype=np.float64)
            basis = np.loadtxt(f)
        uc = WaterModel(latticeparams[:3],latticeparams[3:],watertype)
        if basis.size < 5: # compatibility to old versions
            basis = np.concatenate((basis,[1.]))
        uc.basis = basis
        return uc
    
    @staticmethod
    def fromStr(string):
        with util.StringIO(string) as f:
            next(f) # return line
            nline = next(f).split()
            try:
                latticeparams = np.array(nline,dtype=np.float64)
                watertype = 'step' # default behaviour
            except ValueError:
                watertype = nline[0]
                latticeparams = np.array(next(f).split(),dtype=np.float64)
            try:
                while True:
                    nline = next(f)
                    sline = nline.split()
                    if sline and not 'zpos/frac' in sline:
                        if '+-' in sline:
                            params = re.findall('\(([^)]+)',nline)
                            params_array = np.array([np.array(p.split('+-'),dtype=np.float64) for p in params]).T
                            basis = params_array[0]
                            errors = params_array[1]
                        else:
                            basis = np.array(sline,dtype=np.float64)
                            errors = None
                        if basis.size < 5: # compatibility to old versions
                            basis = np.concatenate((basis,[1.]))
                            if errors is not None:
                                errors = np.concatenate((errors,[np.nan]))
            except StopIteration:
                pass
        uc = WaterModel(latticeparams[:3],latticeparams[3:],watertype)
        uc.basis = basis
        uc.errors = errors
        return uc      


class UnitCell(Crystal):
    
    parameterOrder = "Name   x/frac     y/frac     z/frac     iDW     oDW      occup"
    
    parameterLookup = {'x' : 1, 'y' : 2, 'z' : 3,
                       'iDW' : 4, 'oDW' : 5,
                       'occ' : 6}
    
    # lookuptable for special materials
    # give two functions: 1. for f0 (Q) , 2. lookup for f1 and f2 (E)
    special_formfactors = OrderedDict()
    special_eDensity = OrderedDict()
    special_onset = len(xraydb.get_xraydb().atomic_symbols)
    special_numbers = OrderedDict()
    
    def __init__(self,a,alpha,**keyargs):
        super(UnitCell,self).__init__(a,alpha)
        self.basis = np.array([])
        self.names = []
        self.fitparameters = []
        self.relfitparam = []
        self.fitparlimits = []
        self.relfitlimits = []
        self.basis_0 = None
        self.errors = None
        if 'name' in keyargs:
            self.name = keyargs['name']
        else:
            self.name = "unnamed"
        self.refHKLTransform = np.matrix(np.identity(3))
        self.refRealTransform = np.matrix(np.identity(3))
        self.coherentDomainMatrix = [np.vstack((np.identity(3),np.array([0,0,0]))).T]
        self.coherentDomainOccupancy = [1.]
        self.uc_area = a[0]*a[1]*np.sin(alpha[2]) #unit cell area
        #self.F_uc = UnitCell._F_uc_direct
        #self.F_uc_bulk = UnitCell.F_uc_bulk_direct
            

    def setReferenceUnitCell(self,uc,rotMatrix=np.identity(3)):
        """   
        set reference unit cell. 
        When any F_hkl is called, first hkl will be transformed from the reference 
        lattice into the lattice of this unit cell,
        using equation
        B' * H' = O * B * H, where a ' indicates the entities in the frame of this
        lattice. O is an optional rotation matrix, which describes the rotation of
        this lattice with respect to the reference lattice.
        
        Equally the same applies to the relative coordinates of the respective 
        lattices:
        R' * x' = O * R * x
        
        !!! This has to be checked !!!
        """
        
        self.refRealTransform = self.RealspaceMatrix.I * rotMatrix * uc.RealspaceMatrix 
        self.refHKLTransform =  self.getB().I * rotMatrix *  uc.getB()
        
    
    def addAtom(self,element_or_param, xyz_rel=None, iDW=None, oDW=None, occu=None):
        if xyz_rel is not None:
            atom_no = atomic_number(element_or_param)

            parameters = np.array([atom_no,
                                   *xyz_rel,
                                   iDW,
                                   oDW,
                                   occu],dtype=np.float64)
            self.names.append(element_or_param) 
        else:
            parameters = np.array(element_or_param,dtype=np.float64)
            self.names.append(element_or_param[0]) 
        if self.basis.size > 0:
            self.basis = np.vstack([self.basis,parameters])
        else:
            self.basis = np.array([parameters])
        self.basis_0 = None
        self.errors = None
        return
    
    def insertAtom(self,index,element_or_param, xyz_rel=None, iDW=None, oDW=None, occu=None):
        if xyz_rel is not None:
            atom_no = atomic_number(element_or_param)

            parameters = np.array([atom_no,
                                   *xyz_rel,
                                   iDW,
                                   oDW,
                                   occu],dtype=np.float64)
            self.names.insert(element_or_param,index) 
        else:
            parameters = np.array(element_or_param,dtype=np.float64)
            self.names.insert(element_or_param[0],index) 
        if self.basis.size > 0:
            self.basis = np.insert(self.basis,index,parameters,axis=0)
        else:
            self.basis = np.array([parameters])
        self.basis_0 = None
        self.errors = None
        newfp = []
        for atomindex, parindex in self.fitparameters:
            try:
                if atomindex >= index:
                    atomindex += 1
                newfp.append((atomindex,parindex))
            except TypeError:
                idxarray = []
                for idx in atomindex:
                    if idx >= index:
                        idx += 1
                    idxarray.append(idx)
                newfp.append((tuple(idxarray),parindex))
        self.fitparameters = newfp
        
    # in eV
    def setEnergy(self,E):
        """
        for dispersion and absorption correction
        in eV
        """
        self._E = E
        self.lookupScatteringFactors(E)
        
    def addFitParameter(self,indexarray,limits=[-np.inf,np.inf]):
        """
        """
        if not limits[0] <= np.mean(self.basis[indexarray]) <= limits[1]:
            raise ValueError("start parameter is not within limits! Is not: %s <= %s <= %s" % (limits[0],np.mean(self.basis[indexarray]), limits[1]))  
        self.fitparameters.append(indexarray)
        self.fitparlimits.append(limits)
        return len(self.fitparameters) - 1


    def addRelParameter(self,indexarray,factors,limits=[-np.inf,np.inf]):
        #if not limits[0] <= np.mean(self.basis[indexarray]) <= limits[1]:
        #    raise ValueError("start parameter is not within limits!")  
        self.relfitparam.append((indexarray,factors))
        self.relfitlimits.append(limits)
        return len(self.relfitparam) - 1

    def fixInitialBasis(self):
        self.basis_0 = np.copy(self.basis)

    def getInitialParameters(self):
        if self.basis_0 is None:
            self.basis_0 = np.copy(self.basis)
        
        x0 = np.empty(len(self.fitparameters) + len(self.relfitparam),dtype=np.float)
        
        for i,parray in enumerate(self.fitparameters):
            x0[i] = np.mean(self.basis[parray])
        
        for i,(parray,factor) in enumerate(self.relfitparam,len(self.fitparameters)):
            x0[i] = np.mean( (self.basis[parray] - self.basis_0[parray]) / factor)
        return x0         
        
    def getStartParamAndLimits(self):
        x0 = self.getInitialParameters()
        if x0.size > 0:
            lower, upper = np.array(self.fitparlimits + self.relfitlimits).T
        else:
            lower, upper = np.array([]),  np.array([])
        return x0, lower, upper
        
    def setFitParameters(self,x):
        x_0 = x[:len(self.fitparameters)]
        x_r = x[len(self.fitparameters):]
        for val, parray in zip(x_0,self.fitparameters):
            self.basis[parray] = val
        for val, (parray,factor) in zip(x_r,self.relfitparam):
            self.basis[parray] = factor*val + self.basis_0[parray]

    def setFitErrors(self,errors):
        self.errors = np.full_like(self.basis,np.nan)
        err_0 = errors[:len(self.fitparameters)]
        err_r = errors[len(self.fitparameters):]
        for val, parray in zip(err_0,self.fitparameters):
            self.errors[parray] = val
        for val, (parray,factor) in zip(err_r,self.relfitparam):
            self.errors[parray] = factor*val
        
        

    # returns the structure factor of the unit cell
    # h,k,l have to be 1d arrays 
    def F_uc_bulk(self,h,k,l,atten=0):
        F = np.zeros(h.size,dtype=np.complex128)
        f = np.zeros(h.size,dtype=np.complex128)
        hkl = (self.refHKLTransform * np.vstack((h,k,l))).A
        Q_cart2 = ((self._BMatrix * hkl).A)**2
        Q_para2 = np.sum(Q_cart2[:2],axis=0) #squared!!!
        Q_perp2 = Q_cart2[2] #squared!!!
        Q2 = Q_para2 + Q_perp2 #squared!!!
        
        for i in range(len(self.basis)):
            f[:] =  self.f[i][10] + self.f[i][11] + 1j*self.f[i][12]
            for j in range(5):
                f += self.f[i][j]*np.exp(- self.f[i][j+5]*Q2)
            f *= np.exp(- (self.basis[i][4] * Q_para2 + self.basis[i][5] * Q_perp2)/ (16*np.pi**2))
            f *= self.basis[i][6]
            for mat, weight in zip(self.coherentDomainMatrix,self.coherentDomainOccupancy):
                xyz_rel = np.dot(mat[:,:-1],self.basis[i][1:4] + mat[:,-1].A1).A1
                F += weight * f * np.exp(2j*np.pi * np.sum(hkl.T * xyz_rel,axis=1) ) * math.exp(atten*xyz_rel[2])
        return F/self.volume
    
    # returns the structure factor of the unit cell
    # h,k,l have to be 1d arrays 
    def F_uc_bulk_direct(self,h,k,l,atten=0):
        F = np.zeros(h.size,dtype=np.complex128)
        f = np.zeros(h.size,dtype=np.complex128)
        hkl =  np.vstack((h,k,l))
        Q_cart2 = ((self._BMatrix * hkl).A)**2
        Q_para2 = np.sum(Q_cart2[:2],axis=0) #squared!!!
        Q_perp2 = Q_cart2[2] #squared!!!
        Q2 = Q_para2 + Q_perp2 #squared!!!
        
        for i in range(len(self.basis)):
            f[:] =  self.f[i][10] + self.f[i][11] + 1j*self.f[i][12]
            for j in range(5):
                f += self.f[i][j]*np.exp(- self.f[i][j+5]*Q2)
            f *= np.exp(- (self.basis[i][4] * Q_para2 + self.basis[i][5] * Q_perp2)/ (16*np.pi**2))
            f *= self.basis[i][6]
            for mat, weight in zip(self.coherentDomainMatrix,self.coherentDomainOccupancy):
                xyz_rel = np.dot(mat[:,:-1],self.basis[i][1:4] + mat[:,-1].A1).A1
                F += weight * f * np.exp(2j*np.pi * np.sum(hkl.T * xyz_rel,axis=1) ) * math.exp(atten*xyz_rel[2])
        return F/self.volume
    
    def F_uc(self,h,k,l):
        F = np.zeros(h.size,dtype=np.complex128)
        f = np.zeros(h.size,dtype=np.complex128)
        hkl = (self.refHKLTransform * np.vstack((h,k,l))).A
        Q_cart2 = ((self._BMatrix * hkl).A)**2
        Q_para2 = np.sum(Q_cart2[:2],axis=0) #squared!!!
        Q_perp2 = Q_cart2[2] #squared!!!
        Q2 = Q_para2 + Q_perp2 #squared!!!
        
        for i,name in zip(range(len(self.basis)),self.names):
            if name in UnitCell.special_formfactors:
                f[:] = UnitCell.special_formfactors[name][0](np.sqrt(Q2)) + self.f[i][11] + 1j*self.f[i][12]
            else:
                f[:] =  self.f[i][10] + self.f[i][11] + 1j*self.f[i][12]
                for j in range(5):
                    f += self.f[i][j]*np.exp(- self.f[i][j+5]*Q2)
            f *= np.exp(- (self.basis[i][4] * Q_para2 + self.basis[i][5] * Q_perp2)/ (16*np.pi**2))
            f *= self.basis[i][6]
            for mat, weight in zip(self.coherentDomainMatrix,self.coherentDomainOccupancy):
                xyz_rel = np.dot(mat[:,:-1],self.basis[i][1:4] + mat[:,-1].A1).A1
                F += weight * f * np.exp(2j*np.pi * np.sum(hkl.T * xyz_rel,axis=1) )
        return F/self.volume
    
    def F_bulk(self,h,k,l,atten=0):
        hkl = (self.refHKLTransform * np.vstack((h,k,l))).A
        Fuc = self.F_uc_bulk_direct(*hkl,atten)
        return Fuc/(1- np.exp(- 2j*np.pi * l - atten ))
    
    SQRT2pi = np.sqrt(2*np.pi)
    def zDensity_G(self, z, h,k):
        """
        calculates h,k-th Fourier component of the electron density of the unit cell 
        i.e. 0,0-th component is the commonly used z-projected electron density
        
        The density is normalized to the surface area of the unit cell

        Parameters
        ----------
        z : 1-d array
            z coordinates in Angstrom, should be equidestant 
            and monotonally increasing to avoid numerical issues with convolutions
        h : float
            h-th component index
        k : float
            k-th component index

        Returns
        -------
        1d- array complex128
            complex h,k-th Fourier component of the electron density 
            in electrons/Angstrom**3
            calculate the absolute value to get the electron density

        """
        Qpara2 = np.sum(np.dot(self._BMatrix , np.array([h,k,0])).A**2)
        a,alpha,_,_ = self.getLatticeParameters()
        #uc_area = a[0]*a[1]*np.sin(alpha[2])
        #dispersion = self.f[i][10] + self.f[i][11] + 1j*self.f[i][12]
        zstep = np.diff(z)
        zstep_mean = np.mean(zstep)
        if not np.allclose(zstep,zstep_mean):
            warnings.warn("zDensity: z stepsize is not equal in given z array."
                          "This will result in numerical errors in electron density calculation!")
        
        rho = np.zeros_like(z,dtype=np.complex128)
        rho_i = np.empty_like(z,dtype=np.complex128)
        
        for i,name in zip(range(len(self.basis)),self.names):
            deltaZ2i = self.basis[i][5]/(8*(np.pi)**2)
            deltaPara2i = self.basis[i][4]/(8*(np.pi)**2)
            z_i = self.basis[i][3]*self._a[2]
            y_i_frac = self.basis[i][2]
            x_i_frac = self.basis[i][1]
            if name in UnitCell.special_eDensity:
                rho_i[:] = (self.f[i][11] + 1j*self.f[i][12])/ (np.sqrt(2*np.pi*deltaZ2i))
                rho_i *= np.exp( -0.5 *( deltaPara2i * Qpara2 +  ((z- z_i)**2)/deltaZ2i ))
                rho_i += gaussian_filter1d(UnitCell.special_eDensity[name](z- z_i)*np.exp( -0.5 *( deltaPara2i * Qpara2)),np.sqrt(deltaZ2i)/ zstep_mean)
            else:                
                rho_i[:] = (self.f[i][10] + self.f[i][11] + 1j*self.f[i][12]) / (np.sqrt(2*np.pi*deltaZ2i))
                rho_i *= np.exp( -0.5 *( deltaPara2i * Qpara2 +  ((z- z_i)**2)/deltaZ2i ))
                for j in range(5):
                    exp_dpara = self.f[i][j+5] + 0.5*deltaPara2i
                    exp_dz = self.f[i][j+5] + 0.5*deltaZ2i
                    rho_i += (self.f[i][j]/(np.sqrt(4.*np.pi*exp_dz))) * np.exp(- exp_dpara * Qpara2  - (((z- z_i)**2)/(4*exp_dz))  )
                    
            rho_i *= np.exp(-2j*np.pi*(h*x_i_frac + k*y_i_frac + (z- z_i)/self._a[2]))
            rho += rho_i*self.basis[i][6]
        
        return rho/self.uc_area
    
    
    def zDensity_G_asbulk(self, z, h,k, noUC=30):
        Qpara2 = np.sum(np.dot(self._BMatrix , np.array([h,k,0])).A**2)
        a,alpha,_,_ = self.getLatticeParameters()
        #uc_area = a[0]*a[1]*np.sin(alpha[2])
        #dispersion = self.f[i][10] + self.f[i][11] + 1j*self.f[i][12]
        rho = np.zeros_like(z,dtype=np.complex128)
        rho_i = np.empty_like(z,dtype=np.complex128)
        for no in range(noUC):
            noA = no*self._a[2]
            for i in range(len(self.basis)):
                deltaZ2i = self.basis[i][5]/(8*(np.pi)**2)
                deltaPara2i = self.basis[i][4]/(8*(np.pi)**2)
                z_i = self.basis[i][3]*self._a[2]
                y_i_frac = self.basis[i][2]
                x_i_frac = self.basis[i][1]
                
                rho_i[:] = (self.f[i][10] + self.f[i][11] + 1j*self.f[i][12]) / (np.sqrt(2*np.pi*deltaZ2i))
                rho_i *= np.exp( -0.5 *( deltaPara2i * Qpara2 +  ((z- z_i + noA)**2)/deltaZ2i ))
                        
                
                for j in range(5):
                    exp_dpara = self.f[i][j+5] + 0.5*deltaPara2i
                    exp_dz = self.f[i][j+5] + 0.5*deltaZ2i
                    rho_i += (self.f[i][j]/(np.sqrt(4.*np.pi*exp_dz))) * np.exp(- exp_dpara * Qpara2  - (((z- z_i + noA)**2)/(4*exp_dz))  )
                    
                rho_i *= np.exp(-2j*np.pi*(h*x_i_frac + k*y_i_frac + (z- z_i + noA)/self._a[2]))
                rho += rho_i*self.basis[i][6]
        
        return rho/self.uc_area
    
    def lookupScatteringFactors(self,E):
        self.f = np.empty((self.basis.shape[0],13),dtype=np.float64)
        for i,name in enumerate(self.names):
            if name in UnitCell.special_formfactors:
                self.f[i,:11] = 0.
                self.f[i,11:] = UnitCell.special_formfactors[name][1](E)
            else:
                self.f[i,:11] = readWaasmaier(name)
                self.f[i,11:] = readDispersion(name,E)
        return
    
    def plot3d(self,ucx=1,ucy=1,ucz=1,dwon=False,occuon=False,figure=None,translate=np.array([0.,0.,0.]), **keyargs):
        try:
            from mayavi import mlab
        except ImportError:
            warnings.warn("can not import mayavi: 3D plotting not supported")
            return

        if figure is None:
            figure = mlab.figure()
            
        elcolors = keyargs.get('color')
        if elcolors is None:
            elcolors = [elements.rgb(atomic_number(name)) for name in self.names]
        elif isinstance(elcolors ,tuple):
            elcolors = [elcolors for params in self.basis]
            
        resolution = keyargs.get('resolution')
        if resolution is None:
            resolution = 25

        for i,params in enumerate(self.basis):
            
            radius = elements.ELEMENTS[atomic_number(self.names[i])].covrad*2
            #elcolor_c = keyargs.get('color')
            #if elcolor_c is None:
            #    elcolor_c = elements.rgb(int(params[0]))
            if occuon:
                occup = params[6]
            else:
                occup = 1
            x,y,z = params[1:4]
            positions = np.empty((np.abs(ucx*ucy*ucz),3))
            iDW, oDW = DWtoDisorder(params[4:6])
            signx = np.sign(ucx)
            signy = np.sign(ucy)
            signz = np.sign(ucz)
            no = 0
            for xno in range(1,abs(ucx)+1):
                for yno in range(1,abs(ucy)+1):
                    for zno in range(1,abs(ucz)+1):
                        positions[no] = np.array([x + signx*(xno-1),y + signy*(yno-1),z+ signz*(zno-1)]) + translate
                        if random.random() > occup:
                            positions[no] = np.nan
                        no += 1
            position_cart = (self.RealspaceMatrix*np.matrix(positions).T).A
            sigmas = np.empty_like(position_cart)
            sigmas[:2] = iDW
            sigmas[2] = oDW
            
            if dwon:
                position_cart = np.random.normal(position_cart,sigmas)
            mlab.points3d(*position_cart,scale_factor=radius,color=elcolors[i],resolution=resolution,figure=figure)
            
        return figure
    
    def pos_cart(self,atomNo):
        return (self.RealspaceMatrix*(np.matrix(self.basis[atomNo][1:4])).T).A
        
    def pos_cart_error(self,atomNo):
        xyz = self.pos_cart(atomNo)
        realMatrix = np.matrix(self.RealspaceMatrix.A**2)
        error_zeros = np.nan_to_num(self.errors,0)
        xyz_error = np.sqrt((realMatrix * (np.matrix(error_zeros[atomNo][1:4]**2)).T).A)
        return xyz, xyz_error
        
        
    def distanceVectorAtom(self,atomNo1,atomNo2):
        return self.pos_cart(atomNo1) - self.pos_cart(atomNo2)
    
    def bondingLength(self,xyz_atom):
        pos_atom_cart = self.RealspaceMatrix*(np.matrix(xyz_atom)).T
        blength = np.empty(self.basis.shape[0],dtype=np.float64)
        for i, params in enumerate(self.basis):
            #occup = params[6]
            xyz = params[1:4]
            pos_cart = self.RealspaceMatrix*(np.matrix(xyz)).T
            #iDW, oDW = DWtoDisorder(params[4:6])
            blength[i] = LA.norm(pos_atom_cart - pos_cart)
        return blength
            
    def disorder_error(self,atomNo):
        dr = np.sqrt(self.basis[atomNo][4]/(8*np.pi**2))
        dz = np.sqrt(self.basis[atomNo][5]/(8*np.pi**2))
        errdr = 0.5*dr*(self.errors[atomNo][4]/self.basis[atomNo][4])
        errdz = 0.5*dz*(self.errors[atomNo][5]/self.basis[atomNo][5])
        return dr, dz, errdr, errdz
        
    
    def __repr__(self):
        repr_super = super(UnitCell,self).__repr__()
        return "%s\n%i atoms in unit cell" % (repr_super,len(self.names))
        
    def atomToStr(self,no,showErrors=True):
        param = self.basis[no][1:]
        name = self.names[no]
        if (self.errors is not None) and showErrors:
            err = self.errors[no][1:]
            l = []
            for t in zip(param,err):
                [l.append(ti) for ti in t]
            return "%s  (%.5f +- %.5f)  (%.5f +- %.5f)  (%.5f +- %.5f)  (%.4f +- %.4f)  (%.4f +- %.4f)  (%.4f +- %.4f)" % (name,*l)
        else:
            return "%s     %.5f     %.5f     %.5f  %.4f  %.4f  %.4f" % (name,*param)
        
    def domainsToStr(self):
        s = ""
        for mat, occu in zip(self.coherentDomainMatrix,self.coherentDomainOccupancy):
            matT = np.concatenate((mat[:,:-1],mat[:,-1].T),axis=0)
            s += ("Coherent %.5f   " % occu) + np.array2string(np.array(matT).flatten(), formatter={'float_kind':lambda x: "%.5f" % x},max_line_width=100000)[1:-1] + "\n"
        return s
            
    def latticeRODStr(self):
        a,alpha,_,_ = self.getLatticeParameters()
        return "%.4f %.4f %.4f %.4f %.4f %.4f" % (*a,*np.rad2deg(alpha))
    
    
    def parameterStr(self):
        st = ""
        for i in range(len(self.names)):
            st += str(i).zfill(2) + "  " + self.atomToStr(i) + "\n"
        return st
    
    def parameterStrRod(self):
        st = ""
        for i in range(len(self.names)):
            st += self.atomToStr(i,False) + "\n"
        return st
            
    def __str__(self):
        st = repr(self) + "\n"
        st += "id  " + UnitCell.parameterOrder + "\n"
        return st + self.parameterStr()
            
    def writeSURfile(self,filename):
        with open(filename,'w') as f:
            st = "return\n" + self.domainsToStr() + self.latticeRODStr() + "\n" + self.parameterStrRod()
            f.write(st)
            
    def toRODStr(self):
        return "return\n" + self.domainsToStr() + self.latticeRODStr() + "\n" + self.parameterStrRod()
    
    def toStr(self):
        return "return\n" + self.domainsToStr() + self.latticeRODStr() + "\n" + UnitCell.parameterOrder + "\n" + self.parameterStr()
    
    @staticmethod
    def fromSURfile(surfile):
        with open(surfile,'r') as f:
            next(f) # return line
            line = next(f).split()
            domainoccu = []
            domainmatrix = []
            while(line[0] == 'Coherent'):
                domainoccu.append(float(line[1]))
                mat = np.matrix(line[2:],dtype=np.float64)
                if mat.size > 10:
                    domainmatrix.append(np.concatenate( (mat[:,:-3].reshape((3,3)),mat[:,-3:].T),axis=1))
                else:
                    domainmatrix.append(np.concatenate((mat.reshape((3,3)), np.matrix([0,0,0]).T ),axis=1 ) )
                line = next(f).split()
            if not domainoccu:
                domainoccu.append(1.)
                domainmatrix.append(np.matrix(np.vstack((np.identity(3),np.array([0,0,0]))).T))
            latticeparams = np.array(line,dtype=np.float64)
            names = []
            basis = []
            for l in f:
                line = l.split()
                if line:
                    names.append(line[0])
                    basis.append(np.concatenate(([atomic_number(line[0])], np.array(line[1:],dtype=np.float64) )))
        uc = UnitCell(latticeparams[:3],latticeparams[3:])
        uc.coherentDomainMatrix = domainmatrix
        uc.coherentDomainOccupancy = domainoccu
        if len(basis) >= 2:
            basis = np.vstack(basis)
            if basis.shape[1] == 7:
                uc.basis = basis
                uc.names = names
                return uc
            else:
                raise ValueError("wrong number of atomic parameters. read basis is: {}".format(basis))
        elif basis[0].size == 7:
            uc.basis = np.array([basis])
            uc.names = names
            return uc
        else:
            raise ValueError("wrong number of atomic parameters. read basis is: {}".format(basis))
    
    @staticmethod
    def fromStr(string):
        xprfile = False
        with util.StringIO(string) as f:
            next(f) # return line
            line = next(f).split()
            domainoccu = []
            domainmatrix = []
            while(line[0] == 'Coherent'):
                domainoccu.append(float(line[1]))
                mat = np.matrix(line[2:],dtype=np.float64)
                if mat.size > 10:
                    domainmatrix.append(np.concatenate( (mat[:,:-3].reshape((3,3)),mat[:,-3:].T),axis=1))
                else:
                    domainmatrix.append(np.concatenate((mat.reshape((3,3)), np.matrix([0,0,0]).T ),axis=1 ) )
                line = next(f).split()
            if not domainoccu:
                domainoccu.append(1.)
                domainmatrix.append(np.matrix(np.vstack((np.identity(3),np.array([0,0,0]))).T))
            latticeparams = np.array(line,dtype=np.float64)
            names = []
            basis = []
            errors = []
            indices = []
            statistics = dict()
            for l in f:
                line = l.split()
                if line:
                    if 'Name' in line or '=' in line: # parameter or statistics line
                        if '=' in line:
                            try:
                                splitted = [n.split(',') for n in l.split('=')]
                                splitted = [item for sublist in splitted for item in sublist]
                                for i in range(0,len(splitted),2):
                                    statistics[splitted[i].strip()] = float(splitted[i+1])
                            except Exception:
                                print("Cannot read statistics string: %s" % l)
                        continue
                    if '+-' in line:
                        xprfile = True
                        indices.append(int(line[0]))
                        names.append(line[1])
                        params = re.findall('\(([^)]+)',l)
                        params_array = np.array([np.array(p.split('+-'),dtype=np.float64) for p in params]).T
                        basis.append(np.concatenate(([atomic_number(line[1])], params_array[0] )))
                        errors.append(np.concatenate(([atomic_number(line[1])], params_array[1] )))
                    else:
                        names.append(line[0])
                        basis.append(np.concatenate(([np.nan], np.array(line[1:],dtype=np.float64) )))
            basis = np.vstack(basis)
            if xprfile:
                errors = np.vstack(errors)
        uc = UnitCell(latticeparams[:3],latticeparams[3:])
        uc.coherentDomainMatrix = domainmatrix
        uc.coherentDomainOccupancy = domainoccu
        uc.statistics = statistics
        if len(basis.shape) >= 2:
            basis = np.vstack(basis)
            if basis.shape[1] == 7:
                uc.basis = basis
                if xprfile:
                    uc.errors = errors
                uc.names = names
                return uc
            else:
                raise ValueError("wrong number of atomic parameters. read basis is: {}".format(basis))
        elif basis[0].size == 7:
            uc.basis = np.array([basis])
            if xprfile:
                uc.errors = np.array([errors])
            uc.names = names
            return uc
        else:
            raise ValueError("wrong number of atomic parameters. read basis is: {}".format(basis))

        
    
    #ugly!!!! please redo this for faster file reads!
    @staticmethod
    def fromBULfile(bulfile,DW=0.):
        with open(bulfile,'r') as f:
            next(f) # return line
            line = next(f).split()
            domainoccu = []
            domainmatrix = []
            while(line[0] == 'Coherent'):
                domainoccu.append(float(line[1]))
                mat = np.matrix(line[2:],dtype=np.float64)
                if mat.size > 10:
                    domainmatrix.append(np.concatenate( (mat[:,:-3].reshape((3,3)),mat[:,-3:].T),axis=1))
                else:
                    domainmatrix.append(np.concatenate((mat.reshape((3,3)), np.matrix([0,0,0]).T ),axis=1 ) )
                line = next(f).split()
            if not domainoccu:
                domainoccu.append(1.)
                domainmatrix.append(np.matrix(np.vstack((np.identity(3),np.array([0,0,0]))).T))
            latticeparams = np.array(line,dtype=np.float64)
            #basis = np.loadtxt(f,converters={0 : lambda x :  atomic_number(x.decode("utf-8")) })
            #f.seek(basisstartline)
            names = []
            basis = []
            for l in f:
                line = l.split()
                if line:
                    names.append(line[0])
                    basis.append(np.concatenate(([atomic_number(line[0])], np.array(line[1:],dtype=np.float64) )))
            basis = np.vstack(basis)
        uc = UnitCell(latticeparams[:3],latticeparams[3:])
        uc.coherentDomainMatrix = domainmatrix
        uc.coherentDomainOccupancy = domainoccu
        if len(basis.shape) >= 2:
            basis = np.vstack(basis)
            if basis.shape[1] == 7:
                for i,b in enumerate(basis):
                    uc.addAtom(names[i],b[1:4],DW,DW,1)
                return uc
            else:
                raise ValueError("wrong number of atomic parameters. read basis is: {}".format(basis))
        elif basis[0].size == 7:
            uc.basis = np.array([basis])
            for i,b in enumerate(basis):
                uc.addAtom(names[i],b[1:4],DW,DW,1)
            return uc
        else:
            raise ValueError("wrong number of atomic parameters. read basis is: {}".format(basis))
   
        
    
SQRT2pi = np.sqrt(2*np.pi)
def _eDensity_old(z ,element, deltaZ, Qpara, deltaPara, E):
    atom_parameters = xraydb.read_Waasmaier(element)
    atom_parameters['exponents'] /= (4*np.pi)**2
    dispersion = xraydb.f1_chantler(element,E) + 1j*xraydb.f2_chantler(element,E)
    
    rho_c = ((atom_parameters['c'] + dispersion) / (SQRT2pi*deltaZ)) * \
            np.exp( -0.5 *( (deltaPara**2) * (Qpara**2) +  (z**2)/(deltaZ**2) ))
            
    rho_0 = np.zeros_like(z)
    for a,b in zip(atom_parameters['scale'],atom_parameters['exponents']):
        exp_dpara = b + 0.5*(deltaPara**2)
        exp_dz = b + 0.5*(deltaZ**2)
        rho_0 += (a/(np.sqrt(4.*np.pi*exp_dz))) * np.exp(- exp_dpara * (Qpara**2)  - ((z**2)/(4*exp_dz))  )
        #rho_0 += (1./(np.sqrt(4.*np.pi*exp_dz))) * np.exp(- ((z**2)/(4*exp_dz))  )
    
    return rho_c + rho_0

def DWtoDisorder(dw):
    return np.sqrt(dw/(8*np.pi**2))
        
        
# returns a,b,c for Q = 4pi/lambda * sin th (instead of s)
def readWaasmaier(element):
    xraydb_t = xraydb.get_xraydb()
    wtab = xraydb_t.tables['Waasmaier']

    row = xraydb_t.query(wtab)
    if isinstance(element, int):
        row = row.filter(wtab.c.atomic_number==element).one()
    else:
        row = row.filter(wtab.c.ion==element.title()).one()
    #if len(row) > 0:
    #    row = row[0]

    c = row.offset
    a = json.loads(row.scale); b = np.array(json.loads(row.exponents)) / (4*np.pi)**2;
    #xraydb_t.close()
    return np.concatenate((a,b,[c]))
        
# incorrect for ions!!
def readDispersion(element,E):
    return xraydb.f1_chantler(atomic_number(element),E), xraydb.f2_chantler(atomic_number(element),E)
        
def atomic_number(elementname):
    if elementname in UnitCell.special_numbers:
        return UnitCell.special_numbers[elementname]
    elif elementname[-1] == '+' or elementname[-1] == '-':
        return xraydb.atomic_number(elementname[:-2])
    else:
        return xraydb.atomic_number(elementname)
    
def estimateDispersionCompound(name,E):
    f1 = 0.; f2 = 0.
    for atom, frac in xraydb.chemparse(name).items():
        f1 += xraydb.f1_chantler(atom, E)*frac
        f2 += xraydb.f2_chantler(atom, E)*frac
    return f1, f2
