import numpy as np
import copy
import warnings
from datautils import util
from datautils.xrayutils.CTRplotutil import ctrfigure, CTR, CTRCollection

from datautils.xrayutils.CTRcalc import UnitCell, SXRDCrystal

from scipy import stats

class CTROptimizer():
    
    def __init__(self,xtal,CTRs):
        self.CTRs = copy.deepcopy(CTRs)
        self.xtal = copy.deepcopy(xtal)
        
    def prepareFit(self):
        self.startp, self.lower_bounds, self.higher_bounds = self.xtal.getStartParamAndLimits()
        self.bounds = (self.lower_bounds, self.higher_bounds)
        for ctr in self.CTRs:
            ctr.invrelerrsqrd_weight = ctr.weight * ctr.err**-2
        
    def get_bounds(self):
        return self.bounds
    
    def get_parameters(self):
        return self.xtal.getInitialParameters()
    
    def weighted_residues2(self,x=None):
        if x is not None:    
            self.xtal.setParameters(x)
        residues = []
        for i,ctr in enumerate(self.CTRs):
            F_theo = np.abs(self.xtal.F(ctr.harr,ctr.karr,ctr.l))
            scale = np.prod(F_theo/ctr.sfI)**(1/ctr.l.size) # scale CTR
            residues.append( (ctr.invrelerrsqrd_weight/scale**2) * ((ctr.sfI*scale - F_theo )**2))
        return np.concatenate(residues)
    
    def residues(self,x=None):
        if x is not None:    
            self.xtal.setParameters(x)
        residues = []
        for i,ctr in enumerate(self.CTRs):
            F_theo = np.abs(self.xtal.F(ctr.harr,ctr.karr,ctr.l))
            scale = np.prod(F_theo/ctr.sfI)**(1/ctr.l.size) # scale CTR
            residues.append(ctr.sfI*scale - F_theo )
        return np.concatenate(residues)
    
    def Rfactor(self,x=None):
        if x is not None:    
            self.xtal.setParameters(x)
        residues = []
        Fobs = []
        for i,ctr in enumerate(self.CTRs):
            F_theo = np.abs(self.xtal.F(ctr.harr,ctr.karr,ctr.l))
            scale = np.prod(F_theo/ctr.sfI)**(1/ctr.l.size) # scale CTR
            residues.append(np.abs(ctr.sfI*scale - F_theo) )
            Fobs.append(np.abs(ctr.sfI*scale))
        residues = np.sum(np.concatenate(residues))
        Fobs = np.sum(np.concatenate(Fobs))
        return residues/Fobs
        
    
    def weighted_residues(self,x=None):
        if x is not None:    
            self.xtal.setParameters(x)
        residues = []
        for i,ctr in enumerate(self.CTRs):
            F_theo = np.abs(self.xtal.F(ctr.harr,ctr.karr,ctr.l))
            scale = np.prod(F_theo/ctr.sfI)**(1/ctr.l.size) # scale CTR
            residues.append( np.sqrt(ctr.weight) * ( (ctr.sfI*scale - F_theo ) / (ctr.err*scale) ))
        return np.concatenate(residues)
    
    def fitness(self,x):
        self.xtal.setParameters(x)
        sumchi2 = 0.
        for i,ctr in enumerate(self.CTRs):
            F_theo = np.abs(self.xtal.F(ctr.harr,ctr.karr,ctr.l))
            scale = np.prod(F_theo/ctr.sfI)**(1/ctr.l.size) # scale CTR
            sumchi2 += np.sum( (ctr.invrelerrsqrd_weight/scale**2) * ((ctr.sfI*scale - F_theo )**2) )
        return [sumchi2]
    
    def statistics(self,x):
        self.xtal.setParameters(x)
        residues2 = self.weighted_residues2()
        
        Rfactor = self.Rfactor()
        
        stat = dict()
        
        #variance = np.concatenate([ctr.err**2 for ctr in self.CTRs])
        #varmat_i = np.diag(1/variance)
        
        chi2_result = np.sum(residues2)
        pvalue = 1 - stats.chi2.cdf(chi2_result,residues2.size - x.size)
        chi2_red = chi2_result/(residues2.size - x.size)
        
        pcov = util.leastsq_covariance(self.weighted_residues,x)

        errors = np.sqrt(np.diag(pcov)*chi2_red)
        self.xtal.setFitErrors(errors)
        self.xtal.setParameters(x)
        
        stat['Chisqr'] = chi2_result
        stat['nodatapoints'] = residues2.size
        stat['Chisqr_red'] = chi2_red
        stat['noparameters'] = x.size
        stat['pvalue'] = pvalue
        stat['Rfactor'] = Rfactor
        stat['covariance'] = pcov
        
        return stat
        
    
    def evaluateStatistics(self,x):
        warnings.warn("usage of evaluateStatistics is deprecated, use CTROptimizer.statistics instead!",DeprecationWarning)
        
        self.xtal.setParameters(x)
        residues2 = self.weighted_residues2()
        
        #variance = np.concatenate([ctr.err**2 for ctr in self.CTRs])
        #varmat_i = np.diag(1/variance)
        
        chi2_result = np.sum(residues2)
    
        pvalue = 1 - stats.chi2.cdf(chi2_result,residues2.size - x.size)
        
        chi2_red = chi2_result/(residues2.size - x.size)
        
        pcov = util.leastsq_covariance(self.residues,x)

        errors = np.sqrt(np.diag(pcov)*chi2_red)
        self.xtal.setFitErrors(errors)
        self.xtal.setParameters(x)
        
        return chi2_result, chi2_red, pvalue, residues2.size
    
    def printStatistics(self,x):
        #chi2_result, chi2_red , pvalue, nodatapoints = self.evaluateStatistics(x)
        stat = self.statistics(x)
        print("Chisqr = %.4f, Chisqr_red = %.4f, R-factor = %.4f ,p-value = %.6f, n_refl = %s" % (stat['Chisqr'],stat['Chisqr_red'],stat['Rfactor'],stat['pvalue'],stat['noparameters']))
        
    
    def get_name(self):
        return "CTR optimizer"
    
    def setCTRPlotSettings(self,lrange,plotsize,**settings):
        self.lrange = lrange
        self.plotsize = plotsize
        self.settings = settings
    
    
    """
    def defaultCTRplotsettings(self):
        self.lrange = [0.,9.]
        self.plotsize = (19,12)
        self.settings = {linestyle='', marker='.',color='black',zorder=2,elinewidth=0.5,capsize=1.,'markersize' : 2.}
    
    def plotParametersetCTR(self,x,lrange=[0.,9.],plotsize=(19,12)):
        self.xtal.setParameters(x)
        fitCTRs = self.CTRs.generateCollectionFromXtal(self.xtal,1000,lrange)
        
        ctroverviewfig = ctrfigure(figsize=plotsize)
        
        self.CTRs.setPlotSettings(linestyle='', marker='.',color='black',zorder=2,elinewidth=0.5,capsize=1.,markersize=2.)
        self.CTRs.setAllToDefaultID()
        ctroverviewfig.addCollection(self.CTRs)
        #ideal_rods.setPlotSettings(linestyle='-', marker='',color=(0.7,0.7,0.7,1.),zorder=1,linewidth=4)
        #ideal_rods.setAllToDefaultID()
        #ctroverviewfig.addCollection(ideal_rods)
        fitCTRs.setPlotSettings(linestyle='-', marker='',color='red',zorder=3)
        fitCTRs.setAllToDefaultID()
        ctroverviewfig.addCollection(fitCTRs)
        
        ctroverviewfig.settings(wspace=0.02,hspace=0.05,ylabels='|$F$| / a.u.',ylim=[1e-1,1e1],xlim=[-0.05,8.05],xlabels="$L$ / r.l.u.") 
        ctroverviewfig.generateCTRplot(2)
        
    """
        