# -*- coding: utf-8 -*-
import os
import matplotlib.pyplot as plt
import matplotlib.figure as mplfig
import scipy.interpolate as interp
#from .. import util
# -*- coding: utf-8 -*-
import numpy as np 
from matplotlib import rc
import matplotlib.pyplot as plt
from matplotlib import colors as mcolors
import math
import json
import warnings
from .CTRcalc import SXRDCrystal

# don't init CTRFigure directly, use ctrfigure instead
class CTRFigure(mplfig.Figure):
    
    def __init__(self,**figargs):
        super(CTRFigure, self).__init__(**figargs)
        self.data = {}
        self.xlabels = "L / r.l.u."
        self.ylabels = "Structure factor / a.u."
        self.xlim = None #[0,5]
        self.ylim = None #[3e1,2e3]
        self.wspace = 0
        self.hspace = 0.05
        
        
    def settings(self,**settings):
        self.__dict__.update(settings)
        
        
    def addCTR(self,ctr,*plotargs,**keyargs):
        if ctr.ctr_id in self.data:
            self.data[ctr.ctr_id].append([ctr,plotargs,keyargs])
        else:
            self.data[ctr.ctr_id] = [[ctr,plotargs,keyargs]]
            
    def addCollection(self,collection):
        for rod in collection:
            self.addCTR(rod,*collection.plotsett,**collection.plotkeyargs)
        
            
    def generateCTRplot(self,cols=2,maxrelerr=None,**keyargs):
        self.clear()
        rows = math.ceil(float(len(self.data))/float(cols))
        ax1 = self.add_subplot( rows,cols,1 )
        [self.add_subplot( rows,cols,i, sharex=ax1, sharey=ax1 ) for i in range(2,rows*cols+1)]
        for i,ctrkey in enumerate(sorted(self.data,key=lambda x : np.sum(np.abs(x[0])))):
            ax = self.axes[i]
            ctrname, ctrid = ctrkey
            
            if 'rodLabelL' in keyargs:
                ctrstr = "( %s %s $L$)" % (str(int(ctrname[0])),str(int(ctrname[1])) )
            else:
                ctrstr = "( %s %s )" % (str(int(ctrname[0])),str(int(ctrname[1])) )
            
            if 'rodlabel' in keyargs:
                if keyargs['rodlabel'] == 'bottom':
                    ax.text(0.02,0.02,ctrstr,horizontalalignment='left',
                        verticalalignment='bottom' ,size=15,transform=ax.transAxes, weight='bold')
            else:
                ax.text(0.05,0.95,ctrstr,horizontalalignment='left',
                        verticalalignment='top' ,size=15,transform=ax.transAxes, weight='bold')
            for ctr,plotargs,keyargs_c in self.data[ctrkey]:
                if ctr.isWithError:
                    if maxrelerr is not None:
                        relerr = ctr.err / ctr.sfI
                        mask = relerr > maxrelerr
                        ax.errorbar(ctr.l[~mask],ctr.sfI[~mask],yerr=ctr.err[~mask],**keyargs_c)                        
                        if 'plotcaps' in keyargs:
                            errbar = ax.errorbar(ctr.l[mask],ctr.sfI[mask],yerr=ctr.err[mask],**keyargs_c)
                            _,_,barlinecols = errbar.lines
                            barlinecols[0].set_visible(False)
                    else:
                        ax.errorbar(ctr.l,ctr.sfI,yerr=ctr.err,**keyargs_c) # ,fmt='.',elinewidth=0.5,capsize=1. ,errorevery=1,color='k',zorder=1
                else:
                    ax.plot(ctr.l,ctr.sfI,*plotargs,**keyargs_c)
            if self.data[ctrkey][0][0].difference:    
                ax.set_yscale('linear')
            else:
                ax.set_yscale('log')
                
            if self.xlim is not None:
                ax.set_xlim(self.xlim)
            if self.ylim is not None:
                ax.set_ylim(self.ylim) 
            #ax.grid(True,'both','x')
        
        for ax in self.axes:
            [i.set_linewidth(1.5) for i in ax.spines.values()]
            ax.get_xaxis().set_tick_params(direction='in', which='both',bottom='on', top='on',width=2,length=3)
            ax.get_yaxis().set_tick_params(direction='in', which='both', left='on',right='on',width=2,length=3)
            ax.get_xaxis().set_tick_params(direction='in', which='major',bottom='on', top='on',width=2,length=6)
            ax.get_yaxis().set_tick_params(direction='in', which='major', left='on',right='on',width=2,length=6)
            
        axes = np.reshape(self.axes,(rows,cols))
        plt.setp([a.get_xticklabels() for a in axes[:-1, :].flat], visible=False)
        plt.setp([a.get_yticklabels() for a in axes[:, 1:].flat], visible=False)
        [a.set_ylabel(self.ylabels) for a in axes[:, 0]]
        [a.set_xlabel(self.xlabels) for a in axes[-1, :]]
        
        self.tight_layout()
        self.subplots_adjust(wspace=self.wspace,hspace=self.hspace)
        #fig.savefig("%s/%s_%s.%s" % (fitfolder,name,str(settingsnumber).zfill(4),"pdf"))
        
        
                
class CTR(object):
    
    ctrtype = 100
    
    def __init__(self,hk,l=None,sfI=None,err=None,phi=None,**keyargs):
        self.hk = tuple(hk)
        h,k = hk
        self.l = l
        self.sfI = sfI
        self.err = err
        self.ctrtype = CTR.ctrtype
        CTR.ctrtype += 1
        self.withErr = True
        self.phi = phi
        self.difference = False
        self.weight = 1
        if 'name' in keyargs:
            self.name = keyargs['name']
        else:
            self.name = 'default'
        self.harr = np.full_like(self.l,h)
        self.karr = np.full_like(self.l,k)
        
    def getPlotLabel(self):
        if self.plotlabel in self:
            return self.plotlabel
        else:
            return self.name
            
        
    # in degrees
    def setPhase(self,phi):
        self.phi = phi
    
    def setWithError(self,err):
        self.withErr = err
    
    @property
    def isWithPhase(self):
        if not isinstance(self.phi,np.ndarray):
            return False
        else:
            return True
        
    @property
    def isWithError(self):
        if not isinstance(self.err,np.ndarray):
            return False
        return self.withErr
    
    def getComplexSF(self):
        if not self.isWithPhase:
            raise Exception("%s:\nNo phase informaion available." % repr(self))
        return self.sfI*np.exp(1j*np.deg2rad(self.phi))
    
    @property
    def ctr_id(self):
        return tuple(np.around(self.hk,2)),self.ctrtype
    
    def setToDefaultID(self):
        self.ctrtype = 0
    
    def generateDifference(self,other):
        otherinter = interp.interp1d(other.l,other.sfI)
        self.sfI -= otherinter(self.l)
        self.difference = True
        return self
    
    def meanSF(self,lowerL,upperL):
        upper = np.argmin(np.abs(self.l - upperL))
        lower = np.argmin(np.abs(self.l - lowerL))
        return np.mean(self.sfI[lower:upper])
        
    def __imul__(self,valOrArray):
        self.sfI *= valOrArray
        if isinstance(self.err,np.ndarray):
            self.err *= valOrArray
        return self
    
    def __iadd__(self,valOrArray):
        self.sfI += valOrArray
        #if isinstance(self.err,np.ndarray):
        #    self.err += valOrArray
        return self
    
    def cut(self,lower,upper):
        self.l = self.l[lower:upper]
        self.harr = self.harr[lower:upper]
        self.karr = self.karr[lower:upper]
        self.sfI = self.sfI[lower:upper]
        if self.isWithError:
            self.err = self.err[lower:upper]
        if self.isWithPhase:
            self.phi = self.phi[lower:upper]
            
    def cutToL(self,lowerL,upperL):
        upper = np.argmin(np.abs(self.l - upperL))
        lower = np.argmin(np.abs(self.l - lowerL))
        self.cut(lower,upper)
        
        
    def cutToROIfile(self,jsonfile):
        with open(jsonfile,'r') as f:
            udict = json.load(f)
        
        rois = udict['ROI']['roidict']
        del rois['ICR']
        mask = np.zeros_like(self.l,dtype=np.bool)
        
        for roikey in rois:
            fr = rois[roikey]['from']
            to = rois[roikey]['to']
            lower = np.argmin(np.abs(self.l - fr))
            upper = np.argmin(np.abs(self.l - to))
            mask[lower:upper] = 1.
        self.l = self.l[mask]
        self.harr = self.harr[mask]
        self.karr = self.karr[mask]
        self.sfI = self.sfI[mask]
        
        if self.isWithError:
            self.err = self.err[mask]
        if self.isWithPhase:
            self.phi = self.phi[mask]
        
    def scaleToXtal(self,xtal):
        F_cryst = np.abs( xtal.F(self.harr,self.karr,self.l))
        self.__imul__(np.prod(F_cryst/self.sfI)**(1/self.l.size))
        
    def toArray(self,mode=None):
        if mode is not None:
            data = np.empty((6,self.l.size))
            data[5] = mode
        else:
            data = np.empty((5,self.l.size))
        data[0] = self.harr
        data[1] = self.karr
        data[2] = self.l
        data[3] = self.sfI
        if self.err is not None: 
            data[4] = self.err
        elif self.phi is not None:
            data[4] = self.phi
        else:
            data[4] = np.nan
            
        return data.T

    #returns a list of CTRs!!!
    @staticmethod
    def fromANAROD(filenameOrArray,RODexport=False):
        warnings.warn("CTR.fromANAROD is deprecated, use CTRCollection.fromANAROD instead!", DeprecationWarning)
        if not isinstance(filenameOrArray,np.ndarray):
            filenameOrArray = np.loadtxt(filenameOrArray,skiprows=1)
        rods = np.unique(filenameOrArray[:,:2],axis=0)
        CTRs = []
        for hk in rods:
            rodmask = np.logical_and(filenameOrArray[:,0] == hk[0],filenameOrArray[:,1] == hk[1])
            rod = filenameOrArray[rodmask]
            l = rod[:,2]
            sfI = rod[:,3]
            if RODexport:
                ctr = CTR(tuple(hk),l,sfI)
                ctr.setPhase(rod[:,4])
                CTRs.append(ctr)
            else:
                err = rod[:,4] if rod.shape[1] > 4 else None
                CTRs.append(CTR(tuple(hk),l,sfI,err))
            
        return CTRCollection(CTRs)
    
    @classmethod
    def fromArray(cls, array,RODexport=False):
        h = array[:,0][0]
        k = array[:,1][0]
        l = array[:,2]
        sfI = array[:,3]
        if RODexport:
            err = None
            phase = array[:,4]
        else:
            err = array[:,4] if array.shape[1] > 4 else None
            phase = None
        return cls([h,k],l,sfI,err,phase)
    
    def millerIdentifier(self):
        h,k = self.hk
        idstr = "%i_%02i_%i_%02i" % (round(h), round((h % 1) * 100),round(k), round((k % 1) * 100))
        return idstr
    
    def rollbackHK(self):
        idstrsp = self.name.split('_')
        h = float(idstrsp[0] + '.' + idstrsp[1])
        k = float(idstrsp[2] + '.' + idstrsp[3])
        self.hk = h,k
        
    
    def __repr__(self):
        #return "<CTR %s ctrtype %s at %016X>" % (tuple(np.around(self.hk,2)), self.ctrtype , id(self))
        return "<CTR<%s> %s ctrtype %s>" % (self.name,tuple(np.around(self.hk,2)), self.ctrtype)
    
    #def toANAROD(filename,toFile=True):
        

class CTRCollection(list):
    def __init__(self,iterableCTRs=[]):
        super(CTRCollection, self).__init__(iterableCTRs)
        self.plotsett = ()
        self.plotkeyargs = {'linestyle':'-', 'marker':'.','color':'k','zorder':1}
        #self._updaterodtypes()
        
        
    def setPlotSettings(self,*settings,**keyargs):
        self.plotsett = settings
        self.plotkeyargs = keyargs
    
    def getReprList(self):
        return [repr(rod) for rod in self]
        
    def getHKList(self):
        return [rod.hk for rod in self]
    
    def setAllToDefaultID(self):
        for rod in self:
            rod.setToDefaultID()
    
    def deleteRod(self,key):
        if isinstance(key,tuple):
            for rod in self:
                if rod.hk == key:
                    self.remove(rod)
            
    def __and__(self,other):
        coll = CTRCollection()
        for rod in self:
            if repr(rod) in other.getReprList():
                coll.append(rod)
        #coll._updaterodtypes()
        return coll
    
    def generateDifferenceCollection(self,other,sortby='repr'):
        coll = CTRCollection()
        if isinstance(other,CTRCollection):
            for rod in self:
                if sortby == 'repr':
                    try:
                        idx = other.getReprList().index(repr(rod))
                    except ValueError:
                        continue
                else:
                    try:
                        idx = other.getHKList().index(rod.hk)
                    except ValueError:
                        continue
                coll.append(rod.generateDifference(other[idx]))
            return coll
        elif isinstance(other,SXRDCrystal):
            for ctr in self:
                l = np.copy(ctr.l)
                h,k = ctr.hk
                harr = np.full_like(l,h)
                karr = np.full_like(l,k)
                F_theo = np.abs(other.F(harr,karr,l))
                scale = np.prod(ctr.sfI/F_theo)**(1/ctr.l.size)
                F_theo *= scale
                diffCTR = CTR((h,k), l, ctr.sfI - F_theo,name=ctr.name)
                diffCTR.difference = True
                coll.append(diffCTR)
            return coll
        else:
            raise NotImplementedError("Can not generate difference collection for type %s" % type(other))
        
    def addScalar(self,scalar):
        for rod in self:
            rod += scalar
        return self

    def generateCollectionFromXtal(self,xtal,samples=None,lrange=None):
        CTRs = CTRCollection()
        if samples is not None or lrange is not None:
            if lrange is not None:
                for ctr in self:
                    h,k = ctr.hk
                    l = np.linspace(lrange[0],lrange[1],samples)
                    h = np.full_like(l,h)
                    k = np.full_like(l,k)
                    F =  xtal.F(h,k,l)
                    CTRs.append( CTR( ctr.hk, l, np.abs(F),phi=np.angle(F)))
                return CTRs
            else:
                for ctr in self:
                    h,k = ctr.hk
                    l = np.linspace(np.amin(ctr.l),np.amax(ctr.l),samples)
                    h = np.full_like(l,h)
                    k = np.full_like(l,k)
                    F =  xtal.F(h,k,l)
                    CTRs.append( CTR(ctr.hk, l, np.abs(F),phi=np.angle(F)))
                return CTRs
        else:
            for ctr in self:
                F =  xtal.F(ctr.harr,ctr.karr,ctr.l)
                CTRs.append( CTR(ctr.hk, ctr.l, np.abs(F),phi=np.angle(F)))
            return CTRs
        
    def toANAROD(self,filename, mode=-3):
        if self.__getitem__(0).isWithError:
            header = "H  K  L  F_HKL  errorF  mode"
        else:
            header = "H  K  L  F_HKL  phi"
            
        data_combined = np.vstack([ctr.toArray(mode) for ctr in self])
        
        np.savetxt(filename,data_combined,header=header,fmt='%.5f')
        
                    
    @staticmethod
    def fromANAROD(filenameOrArray,RODexport=False):
        if not isinstance(filenameOrArray,np.ndarray):
            filenameOrArray = np.loadtxt(filenameOrArray,skiprows=1)
        rods = np.unique(filenameOrArray[:,:2],axis=0)
        CTRs = []
        for hk in rods:
            rodmask = np.logical_and(filenameOrArray[:,0] == hk[0],filenameOrArray[:,1] == hk[1])
            rod = filenameOrArray[rodmask]
            l = rod[:,2]
            sfI = rod[:,3]
            if RODexport:
                ctr = CTR(tuple(hk),l,sfI)
                ctr.setPhase(rod[:,4])
                CTRs.append(ctr)
            else:
                err = rod[:,4] if rod.shape[1] > 4 else None
                CTRs.append(CTR(tuple(hk),l,sfI,err))
            
        return CTRCollection(CTRs)    

    def scaleToXtal(self,xtal):
        for ctr in self:
            ctr.scaleToXtal(xtal)
    
    def __imul__(self,valOrArray):
        for rod in self:
            rod *= valOrArray
        return self

    def __getitem__(self,key):
        if isinstance(key,tuple):
            for rod in self:
                if rod.hk == key:
                    return rod
        return super(CTRCollection, self).__getitem__(key)

     
def ctrfigure(**figargs):
    figargs['FigureClass'] = CTRFigure
    fig = plt.figure(**figargs)
    return fig
    
if __name__ == "__main__":
    fig = ctrfigure(figsize=(12,8))
    ctrlist = CTR.fromANAROD("CTRfit/data_in/0V17/CH4977_0V17_0001.dat")
    [ctr.setWithError(False) for ctr in ctrlist]
    [ctr.setToDefaultID() for ctr in ctrlist]
    [fig.addCTR(r,linestyle='', marker='.',color='k',zorder=1) for r in ctrlist]
    ctrlist = CTR.fromANAROD("CTRfit/data_in/0V47/CH4977_0V47_0001.dat")
    [ctr.setWithError(False) for ctr in ctrlist]
    [ctr.setToDefaultID() for ctr in ctrlist]
    [fig.addCTR(r,linestyle='', marker='.',color='g',zorder=1) for r in ctrlist]
    
    ctrlist = CTR.fromANAROD("CTRfit/data_in/0V72/CH4977_0V72_0001.dat")
    [ctr.setWithError(False) for ctr in ctrlist]
    [ctr.setToDefaultID() for ctr in ctrlist]
    [fig.addCTR(r,linestyle='', marker='.',color='b',zorder=1) for r in ctrlist]
    
    ctrlist = CTR.fromANAROD("CTRfit/data_in/0V02/CH4977_0V02_0001.dat")
    [ctr.setWithError(False) for ctr in ctrlist]
    [ctr.setToDefaultID() for ctr in ctrlist]
    [fig.addCTR(r,linestyle='', marker='.',color='y',zorder=1) for r in ctrlist]
    
    fig.settings(wspace=0.05,hspace=0,ylabels='|F| / a.u.',ylim=[1e1,1e3]) 
    fig.generateCTRplot()
    fig.show()
    

"""
#for i,rod in enumerate(CTRs,1):
for i,ax in enumerate(axes.flat):
    #ax = fig.add_subplot(rows,cols,i)
    if i == len(CTRs):
        break
    roddata = CTRs[i][0]
    fit = CTRs[i][1]

    ax.errorbar(roddata[:,2],roddata[:,3],yerr=roddata[:,4],fmt='.',elinewidth=0.5,capsize=1. ,errorevery=1,color='k',zorder=1)
    if showCalculated:
        ax.plot(fit[:,2],fit[:,3],'r-',zorder=10)
    ax.set_yscale('log')

    ax.text(0.05,0.95,"( %s %s )" % (str(int(roddata[0][0])),str(int(roddata[0][1])) ),horizontalalignment='left',
            verticalalignment='top' ,size=15,transform=ax.transAxes)
    #ax.set_xlabel("L / r.l.u.")
    #ax.set_ylabel(r'$|F_{'+ str(int(roddata[0][0])) + str(int(roddata[0][1]))+ r'L}|$ / a.u.')
    ax.set_xlim([0,5])
    ax.set_ylim([3e0,1e3])
    ax.grid(True,'both','x')
plt.setp([a.get_xticklabels() for a in axes[0, :]], visible=False)
plt.setp([a.get_yticklabels() for a in axes[:, 1]], visible=False)
[a.set_ylabel(r'$|F|$ / a.u.') for a in axes[:, 0]]
[a.set_xlabel("L / r.l.u.") for a in axes[1, :]]

fig.tight_layout()
fig.subplots_adjust(wspace=wspace,hspace=hspace)
fig.savefig("%s/%s_%s.%s" % (fitfolder,name,str(settingsnumber).zfill(4),"pdf"))

"""

    

    
    














