# -*- coding: utf-8 -*-
"""
Version 6: 
    added basic bliss support 

Version 5:
    added cbf support for P3_Image
    added fastscan class for data management of fastscans / fasttake
    

"""
import numpy as np
from PyMca5.PyMcaIO import EdfFile
from PyMca5.PyMcaIO import PilatusCBF
from PyMca5.PyMcaIO import specfile
import scipy.ndimage
import scipy.interpolate
import scipy.optimize
from matplotlib.colors import LinearSegmentedColormap
import datetime
import re
sqrt3 = np.sqrt(3)
import warnings
import fabio
import h5py
import os


class P3_Image:
    def __init__(self, filename,edf=True):
        if edf:
            image = fabio.open(filename)
            #edf = EdfFile.EdfFile(filename, 'r')
            #self.header = edf.GetHeader(0)
            self.header = image.header
            #self.header = edf.GetHeader(0)
            self.motors = dict()
            motor_mne = self.header['motor_mne'].split()
            motor_pos = self.header['motor_pos'].split()
            for i in range(len(motor_mne)):
                self.motors[motor_mne[i]] = float(motor_pos[i])
            self.counters = dict()
            counter_mne = self.header['counter_mne'].split()
            counter_pos = self.header['counter_pos'].split()
            for i in range(len(counter_mne)):
                self.counters[counter_mne[i]] = float(counter_pos[i])
            self.img = image.data.astype(np.float)
            #self.img = np.array(edf.GetData(0), dtype=float)
        else:
            image = fabio.open(filename)
            #cbf = PilatusCBF.PilatusCBF(filename)
            self.header = image.header
            #self.header = cbf.getInfo('') # only detector related information
            self.motors = dict() # not available in cbf
            self.counters = dict() # not available in cbf
            self.img = image.data.astype(np.float)
            #del cbf


class MPX_Image:
    def __init__(self, filename):
        edf = EdfFile.EdfFile(filename, 'r')
        self.header = edf.GetHeader(0)
        self.motors = dict()
        motor_mne = self.header['motor_mne'].split()
        motor_pos = self.header['motor_pos'].split()
        for i in range(len(motor_mne)):
            self.motors[motor_mne[i]] = float(motor_pos[i])
        self.counters = dict()
        counter_mne = self.header['counter_mne'].split()
        counter_pos = self.header['counter_pos'].split()
        for i in range(len(counter_mne)):
            self.counters[counter_mne[i]] = float(counter_pos[i])
        self.img = np.array(edf.GetData(0))


class PE_Image:
    def __init__(self, filename):
        edf = EdfFile.EdfFile(filename, 'r')
        self.header = edf.GetHeader(0)
        self.motors = dict()
        motor_mne = self.header['motor_mne'].split()
        motor_pos = self.header['motor_pos'].split()
        for i in range(len(motor_mne)):
            self.motors[motor_mne[i]] = float(motor_pos[i])
        self.counters = dict()
        counter_mne = self.header['counter_mne'].split()
        counter_pos = self.header['counter_pos'].split()
        for i in range(len(counter_mne)):
            self.counters[counter_mne[i]] = float(counter_pos[i])

        self.img = np.array(edf.GetData(0))
        # detector was mounted rotated by 90deg to the right
        self.img = np.rot90(self.img, 3)
        

# currently only set up for th scans in TOMO session, cbf fileformat
class Fastscan(object):
    def __init__(self,fastscan_specfile, scanno):
        id31_fastscan_spec = specfile.Specfile(fastscan_specfile)
        scan = id31_fastscan_spec[scanno-1]
        filename_line = scan.header('C next_image_file')[0]
        self.filename_base, next_image_name = filename_line.split('/')[-2:]
        
        imagnostr = next_image_name.split('.')[0].split('_')[-1]
        imagnostr_clean = ''.join(list(filter( str.isdigit ,imagnostr)))
        self.directFolder = False
        if imagnostr_clean != imagnostr:
            warnings.warn(("Imageno formatting doesn't follow the fastscan numbering scheme.\n"
                           "Fall back to direct folder selection. (please select folder manually!!)")
                          ,UserWarning)
            self.directFolder = True
            cutposition = next_image_name.find(imagnostr_clean)
            self.filename_base = next_image_name[:cutposition]
        self.first_imageno = int(imagnostr_clean)
        
        #self.first_imageno = int( next_image_name.split('.')[0].split('_')[-1] )
        try:
            self.starttime = datetime.datetime.strptime(scan.header('D')[0][3:]\
                                                    , "%a %b %d %H:%M:%S %Y") 
        except Exception as e:
            warnings.warn("can not read starttime, don't know why yet!. Error: %s" % e)
        self.sr_status = dict(zip(scan.header('UMI0')[0].split()[1:],\
                        re.split(r'\s{2,}', scan.header('UMI1')[0])[1:]))
        scan = id31_fastscan_spec[scanno]
        self.th = np.mean([scan.datacol('th_UpPos'),\
              scan.datacol('th_DownPos')],axis=0)
        self.omega = -1*self.th
        self.exposure_time = scan.datacol('TrigTime')
        commands = scan.command().split()
        self.nopoints = int(commands[-4])
        totaltime = (self.nopoints * float(commands[2]))/float(commands[-3])
        self.time = np.linspace(0,totaltime,self.nopoints)
        self.current = None
        
    def set_th_offset(self,offset):
        self.th += offset
        self.omega = -1*self.th

        
    def set_image_folder(self,path_to_folder):
        #self.filenames = [None]*len(self.th)
        self.filenames = []
        if self.directFolder:
            for i in range(self.th.size):
                self.filenames.append("%s/%s%04i.cbf" % (path_to_folder,\
                    self.filename_base,self.first_imageno+i))
        else:
            for i in range(self.th.size):
                self.filenames.append("%s/%s/%s_%04i.cbf" % (path_to_folder,\
                    self.filename_base,self.filename_base,self.first_imageno+i))
            
    def get_raw_p3_img(self,img):
        return P3_Image(self.filenames[img],False)
    
    # returns single default image!
    def get_raw_img(self,img):
        return P3_Image(self.filenames[img],False)
        
    def get_p3_img(self,img):
        imgdata = P3_Image(self.filenames[img],False)
        imgdata.counters['Time'] = self.time[img]
        imgdata.counters['TrigTime'] = self.exposure_time[img]
        imgdata.counters['th'] = self.th[img]
        imgdata.counters['om'] = self.omega[img]
        if self.current is not None:
            imgdata.counters['srcur'] = self.current[img]
        return imgdata

    # for normalization
    # example: 0.004213 mA/s was ok for 16 bunch, top up mode (Feb 2018) 
    def estimateRingCurrent(self,lossPerSecond):
        self.current = float(self.sr_status['Current']) -\
        lossPerSecond*self.time
        return self.current

    def get_p3_img_recal(self,img):
        imgdata = self.get_p3_img(img)
        imgdata.img = np.multiply(imgdata.img, id31_cal_matrix)
        return imgdata
    
    def __getitem__(self,key):
        return self.get_p3_img(key)
    
    def __len__(self):
        return self.nopoints
        
    
# TODO: change subclassing! this is terrible, but works for the moment
class BlissScan(Fastscan):
    def __init__(self,hdffilepath, scanname):
        filepath,filename = os.path.split(hdffilepath)
        #print(filepath,filename)
        self.scanname = scanname
        _,filename_noext = os.path.split(filepath)
        #filename_noext = filename.split('.')[0]
        self.filename_base = filename_noext #filename_noext[:filename_noext.rfind('_')]
        #print(self.filename_base)
        self.cameras = []
        #self.fixed_th = None
        #self.th = None
        #self.mu = None
        with h5py.File(hdffilepath,'r') as f:
            if scanname not in f:
                raise Exception("Scan name %s does not exist in file" % scanname)
            scangroup = f[scanname]
            command = scangroup['title'][()].decode('utf-8')
            measgroup = scangroup['measurement']
            if command.startswith("cscan"):
                for key in measgroup['th']:
                    try:
                        data = measgroup['th'][key][:]
                        dataname = key.split(':')[1]
                    except ValueError:
                        continue
                    setattr(self,dataname,data)
                    
                self.th = self.th_mean
                self.current = None
                self.nopoints = self.th.size
                self.time = self.time_mean
                self.exposure_time = self.time_delta
                
                if "p3" in measgroup:
                    self.cameras.append('p3')
                    
                if 'mpx' in measgroup:
                    self.cameras.append('mpx')
                
            else:
                for key in measgroup:
                    if key.startswith('group'):
                        scanmotorgroup = measgroup[key]
                        break
                for key in scanmotorgroup:
                    try:
                        data = scanmotorgroup[key][:]
                        dataname = key.split(':')[1]
                    except ValueError:
                        continue
                    
                    setattr(self,dataname,data)
                    
                for key in measgroup['timer']:
                    try:
                        data = measgroup['timer'][key][:]
                        dataname = key.split(':')[1]
                    except ValueError:
                        continue
                    setattr(self,dataname,data)
                    
                positioner_group = measgroup['instrument/positioners']
                if 'th' in positioner_group:
                    self.th = np.full_like(self.srcur,positioner_group['th'][()])
                
                self.current = self.srcur
                self.nopoints = self.srcur.size
                self.time = self.elapsed_time
                self.exposure_time = np.full_like(self.current, float(command.split(' ')[-1]) )
        
                if "p3" in measgroup:
                    self.cameras.append('p3')
                    
                if 'mpx' in measgroup:
                    self.cameras.append('mpx')
                    
                #self.cameras = measgroup['musst'].keys()
            if hasattr(self,"th"):
                self.omega = -1*self.th
                #  To implement:
                #self.starttime = datetime.datetime.strptime(scan.header('D')[0][3:]\
                #                                        , "%a %b %d %H:%M:%S %Y") 

    def set_image_folder(self,path_to_folder):
        #self.filenames = [None]*len(self.th)
        self.filenames = []
        for camera in self.cameras:
            for i in range(len(self)):
                self.filenames.append("%s/%s_%s_%s_%04i.edf.gz" % (path_to_folder,\
                    self.filename_base,camera,self.scanname,i))

    def get_p3_img(self,img):
        imgdata = P3_Image(self.filenames[img],False)
        imgdata.counters['Time'] = self.time[img]
        imgdata.counters['TrigTime'] = self.exposure_time[img]
        if hasattr(self,"th"): 
            imgdata.counters['th'] = self.th[img]
            imgdata.counters['om'] = self.omega[img]
        #if hasattr(self,"fixed_th"):        
        #    imgdata.counters['th'] = self.fixed_th
        #    imgdata.counters['om'] = self.fixed_th*-1
        if hasattr(self,"mu"):       
            imgdata.counters['mu'] = self.mu[img]
        if self.current is not None:
            imgdata.counters['srcur'] = self.current[img]
        return imgdata                    

class Imshow_Formatter(object):
    def __init__(self, im):
        self.im = im

    def __call__(self, x, y):
        y_pos = int(y+0.5)
        x_pos = int(x+0.5)
        img = self.im.get_array()
        if(x_pos > img.shape[0]-1):
            x_pos = img.shape[0]-1
        if(y_pos > img.shape[1]-1):
            y_pos = img.shape[1]-1
        z = img[y_pos, x_pos]
        return 'x={:.01f}, y={:.01f}, z={:d}'.format(x, y, z)


class coords_formatter:
    def __init__(self, img):
        self.img = img

    def format_coord(self, x, y):
        numrows, numcols = self.img.shape
        col = int(x+0.5)
        row = int(y+0.5)
        if col >= 0 and col < numcols and row >= 0 and row < numrows:
            z = self.img[row, col]
            return 'x=%1.4f, y=%1.4f, z=%1.4f' % (x, y, z)
        else:
            return 'x=%1.4f, y=%1.4f' % (x, y)


class Reflection:
    def __init__(self, H, K, L, a_star, b_star, c_star, alpha):
        self.H = H
        self.K = K
        self.L = L
        self.a_star = a_star
        self.b_star = b_star
        self.c_star = c_star
        self.alpha = alpha


id31_img_path = ''
id31_cal_file = ''
id31_cal_matrix = 0


def rad(x):
    return x*np.pi/180.


def deg(x):
    return x*180./np.pi


def get_K_0(E_keV):
    # AA-1
    return 2.*np.pi/12.39854*E_keV

# xy, del-gam conversion for


def get_del(x, cen_x, d):
    return deg(np.arctan((x-cen_x)*0.1720/d))


def get_gam(y, cen_y, d):
    return deg(np.arctan(-(y-cen_y)*0.1720/d))


def get_x(delta, cen_x, d):
    return cen_x + d/0.1720*np.tan(rad(delta))


def get_y(gamma, cen_y, d):
    return cen_y - d/0.1720*np.tan(rad(gamma))


def get_q(k_0, theta, delta, gamma):
    q_x = k_0*(+np.cos(rad(gamma))*np.sin(rad(delta-theta))+np.sin(rad(theta)))
    q_y = k_0*(np.sin(rad(gamma)))
    q_z = k_0*(-np.cos(rad(gamma))*np.cos(rad(delta-theta))+np.cos(rad(theta)))
    return np.array((q_x, q_y, q_z))


def get_q_from_HKL(k_0, H, K, L, a_star, c_star, alpha):
    q = np.zeros(3)
    q[0] = (H+K/2.)*a_star
    q[1] = np.sqrt(3)/2.*K*a_star
    q[2] = L*c_star
    q_Rx = np.cos(rad(-alpha))*q[0]+np.sin(rad(-alpha))*q[1]
    q_Ry = -np.sin(rad(-alpha))*q[0]+np.cos(rad(-alpha))*q[1]

    q[0] = q_Rx
    q[1] = q_Ry

    return q


def get_q_from_HKL_cubic(k_0, H, K, L, a_star, b_star, c_star, alpha):
    q = np.zeros(3)
    q[0] = H*a_star
    q[1] = K*b_star
    q[2] = L*c_star
    q_Rx = np.cos(rad(-alpha))*q[0]+np.sin(rad(-alpha))*q[1]
    q_Ry = -np.sin(rad(-alpha))*q[0]+np.cos(rad(-alpha))*q[1]

    q[0] = q_Rx
    q[1] = q_Ry

    return q


def get_HKL(q, a_star, c_star, alpha):
    q_Rx = np.cos(rad(alpha))*q[0]+np.sin(rad(alpha))*q[1]
    q_Ry = -np.sin(rad(alpha))*q[0]+np.cos(rad(alpha))*q[1]
    H = (q_Rx - q_Ry/sqrt3)/a_star
    K = 2./sqrt3*q_Ry/a_star
    L = q[2]/c_star
    return np.array((H, K, L))


def get_HKL_cubic(q, a_star, b_star, c_star, alpha):
    q_Rx = np.cos(rad(alpha))*q[0]+np.sin(rad(alpha))*q[1]
    q_Ry = -np.sin(rad(alpha))*q[0]+np.cos(rad(alpha))*q[1]
    H = q_Rx/a_star
    K = q_Ry/b_star
    L = q[2]/c_star
    return np.array((H, K, L))


def get_HKL_from_xyth(x, cen_x, y, cen_y, theta, k_0, a_star, c_star, alpha,
                      d):
    delta = get_del(x, cen_x, d)
    gamma = get_gam(y, cen_y, d)
    return get_HKL(get_q(k_0, theta, delta, gamma), a_star, c_star, alpha)


def get_HKL_from_xyth_cubic(x, cen_x, y, cen_y, theta, k_0, a_star, b_star,
                            c_star, alpha, d):
    delta = get_del(x, cen_x, d)
    gamma = get_gam(y, cen_y, d)
    return get_HKL_cubic(get_q(k_0, theta, delta, gamma), a_star, b_star,
                         c_star, alpha)


def _help_func(theta, gamma, q_x, q_z, k_0):
    # all angles in rad
    return (q_z - k_0 * (np.cos(gamma) * np.cos(np.arcsin((q_x / k_0 -
            np.sin(theta) / np.cos(gamma))))-np.cos(theta)))


def get_angles(k_0, H, K, L, a_star, c_star, alpha=0, theta=None):
    q_Rx = a_star*(H+K/2.)
    q_Ry = a_star*(sqrt3/2.*K)
    q_x = np.cos(rad(-alpha))*q_Rx+np.sin(rad(-alpha))*q_Ry
    q_y = -np.sin(rad(-alpha))*q_Rx+np.cos(rad(-alpha))*q_Ry

    th_given = (theta is not None)
    gamma = np.arcsin(q_y/k_0)
    if(not th_given):
        q_z = L*c_star
        A = np.cos(gamma)
        # TODO: this part needs testing
        theta1 = -2*np.arctan((2*q_x/k_0 -np.sqrt(-A**4 +2*A**2*(q_x/k_0)**2 + 2*A**2*(q_z/k_0)**2 + 2*A**2 - (q_x/k_0)**4 - 2*(q_x/k_0)**2*(q_z/k_0)**2 + 2*(q_x/k_0)**2 - (q_z/k_0)**4 + 2*(q_z/k_0)**2 - 1))/(A**2 - (q_x/k_0)**2 - (q_z/k_0)**2 - 2*q_z/k_0 - 1))
        theta2 = -2*np.arctan((2*q_x/k_0 + np.sqrt(-A**4 + 2*A**2*(q_x/k_0)**2 + 2*A**2*(q_z/k_0)**2 + 2*A**2 - (q_x/k_0)**4 - 2*(q_x/k_0)**2*(q_z/k_0)**2 + 2*(q_x/k_0)**2 - (q_z/k_0)**4 + 2*(q_z/k_0)**2 - 1))/(A**2 - (q_x/k_0)**2 - (q_z/k_0)**2 - 2*q_z/k_0 - 1))        
        if(np.abs(theta1) > np.abs(theta2)):
            theta = theta2
        else:
            theta = theta1
    else:
        theta = rad(theta)

    arg = (q_x/k_0-np.sin(theta))/np.cos(gamma)
    delta = np.arcsin(arg) + theta

    if (arg > 1 or arg < -1):
        print('Invalid value ' + str(arg) + ' in arcsin!')
        print('qx = ' + str(q_x) + ', k_0 = ' + str(k_0) + ', theta = ' +
              str(theta) + ', gamma = ' + str(gamma))
        print('H = ' + str(H) + ', K = ' + str(K) + ', L = ' + str(L) +
              ', alpha = ' + str(alpha) + ', theta = ' + str(theta))

    if(th_given):
        q_z = k_0*(-np.cos(gamma)*np.cos(delta-theta)+np.cos(theta))
        LL = q_z/c_star
        return np.array((LL, deg(delta), deg(gamma)))

    return np.array((deg(theta), deg(delta), deg(gamma)))


def get_angles_cubic(k_0, H, K, L, a_star, b_star, c_star, alpha=0,
                     theta=None):
    q_Rx = a_star*H
    q_Ry = b_star*K
    q_x = np.cos(rad(-alpha))*q_Rx+np.sin(rad(-alpha))*q_Ry
    q_y = -np.sin(rad(-alpha))*q_Rx+np.cos(rad(-alpha))*q_Ry

    th_given = (theta is not None)
    gamma = np.arcsin(q_y/k_0)
    if(not th_given):
        q_z = L*c_star
        A = np.cos(gamma)
        # TODO: this part needs testing
        theta1 = -2*np.arctan((2*q_x/k_0 -np.sqrt(-A**4 +2*A**2*(q_x/k_0)**2 + 2*A**2*(q_z/k_0)**2 + 2*A**2 - (q_x/k_0)**4 - 2*(q_x/k_0)**2*(q_z/k_0)**2 + 2*(q_x/k_0)**2 - (q_z/k_0)**4 + 2*(q_z/k_0)**2 - 1))/(A**2 - (q_x/k_0)**2 - (q_z/k_0)**2 - 2*q_z/k_0 - 1))
        theta2 = -2*np.arctan((2*q_x/k_0 + np.sqrt(-A**4 + 2*A**2*(q_x/k_0)**2 + 2*A**2*(q_z/k_0)**2 + 2*A**2 - (q_x/k_0)**4 - 2*(q_x/k_0)**2*(q_z/k_0)**2 + 2*(q_x/k_0)**2 - (q_z/k_0)**4 + 2*(q_z/k_0)**2 - 1))/(A**2 - (q_x/k_0)**2 - (q_z/k_0)**2 - 2*q_z/k_0 - 1))        
        if(np.abs(theta1) > np.abs(theta2)):
            theta = theta2
        else:
            theta = theta1
    else:
        theta = rad(theta)

    arg = (q_x/k_0-np.sin(theta))/np.cos(gamma)
    delta = np.arcsin(arg) + theta

    if (arg > 1 or arg < -1):
        print('Invalid value ' + str(arg) + ' in arcsin!')
        print('qx = ' + str(q_x) + ', k_0 = ' + str(k_0) + ', theta = ' +
              str(theta) + ', gamma = ' + str(gamma))
        print('H = ' + str(H) + ', K = ' + str(K) + ', L = ' + str(L) +
              ', alpha = ' + str(alpha) + ', theta = ' + str(theta))

    if(th_given):
        q_z = k_0*(-np.cos(gamma)*np.cos(delta-theta)+np.cos(theta))
        LL = q_z/c_star
        return np.array((LL, deg(delta), deg(gamma)))

    return np.array((deg(theta), deg(delta), deg(gamma)))


def get_angles_old(k_0, H, K, L, a_star, c_star, alpha, theta=None):
    q_Rx = a_star*(H+K/2.)
    q_Ry = a_star*(sqrt3/2.*K)
    q_x = np.cos(rad(-alpha))*q_Rx+np.sin(rad(-alpha))*q_Ry
    q_y = -np.sin(rad(-alpha))*q_Rx+np.cos(rad(-alpha))*q_Ry

    th_given = (theta is not None)
    gamma = np.arcsin(q_y/k_0)
    if(not th_given):
        q_z = L*c_star
        theta = scipy.optimize.brentq(_help_func, -np.pi*0.48, np.pi*0.48,
                                      (gamma, q_x, q_z, k_0))
    else:
        theta = rad(theta)

    delta = np.arcsin((q_x/k_0-np.sin(theta))/np.cos(gamma))+theta

    if(th_given):
        q_z = k_0*(-np.cos(gamma)*np.cos(delta-theta)+np.cos(theta))
        LL = q_z/c_star
        return np.array((LL, deg(delta), deg(gamma)))

    return np.array((deg(theta), deg(delta), deg(gamma)))


def get_xy_theta(k_0, H, K, L, a_star, c_star, alpha, cen_x, cen_y, d):
    angles = get_angles(k_0, H, K, L, a_star, c_star, alpha)
    x = get_x(angles[1], cen_x, d)
    y = get_y(angles[2], cen_y, d)
    return np.array((x, y, angles[0]))


def get_xy_theta_cubic(k_0, H, K, L, a_star, b_star, c_star, alpha, cen_x,
                       cen_y, d):
    angles = get_angles_cubic(k_0, H, K, L, a_star, b_star, c_star, alpha)
    x = get_x(angles[1], cen_x, d)
    y = get_y(angles[2], cen_y, d)
    return np.array((x, y, angles[0]))


def get_xy_L(k_0, H, K, theta, a_star, c_star, alpha, cen_x, cen_y, d):
    angles = get_angles(k_0, H, K, 0, a_star, c_star, alpha, theta)
    x = get_x(angles[1], cen_x, d)
    y = get_y(angles[2], cen_y, d)
    return np.array((x, y, angles[0]))


def get_xy_L_cubic(k_0, H, K, theta, a_star, b_star, c_star, alpha,
                   cen_x, cen_y, d):
    angles = get_angles_cubic(k_0, H, K, 0, a_star, b_star, c_star, alpha,
                              theta)
    # print(str(angles[1]) + ', ' + str(angles[2]))
    x = get_x(angles[1], cen_x, d)
    y = get_y(angles[2], cen_y, d)
    # print(str(x) + ', ' + str(y))
    return np.array((x, y, angles[0]))


def blue_white_red_cmap(white_region=0.1):
    region_2 = (0.5 - white_region)/2.
    region_3 = 0.5 - white_region/2.
    region_4 = 0.5 + white_region/2.
    region_5 = 1. - (0.5 - white_region)/2.

    cdict = {'red':  ((0.0, 0.0, 0.0),
                      (region_2, 0.0, 0.0),
                      (region_3, 1.0, 1.0),
                      (region_4, 1.0, 1.0),
                      (region_5, 1.0, 1.0),
                      (1.0, 0.5, 0.0)),

             'green': ((0.0, 0.0, 0.0),
                       (region_2, 0.0, 0.0),
                       (region_3, 1.0, 1.0),
                       (region_4, 1.0, 1.0),
                       (region_5, 0.0, 0.0),
                       (1.0, 0.0, 0.0)),

             'blue':  ((0.0, 0.0, 0.5),
                       (region_2, 1.0, 1.0),
                       (region_3, 1.0, 1.0),
                       (region_4, 1.0, 1.0),
                       (region_5, 0.0, 0.0),
                       (1.0, 0.0, 0.0))
             }

    return LinearSegmentedColormap('BlueRed', cdict)


def red_white_cmap(white_region=0.1):
    region_2 = (0.5 - white_region)/2.
    region_3 = 0.5 - white_region/2.
    region_4 = 0.5 + white_region/2.
    region_5 = 1. - (0.5 - white_region)/2.

    cdict = {'red':  ((0.0, 1.0, 1.0),
                      (region_2, 1.0, 1.0),
                      (region_3, 1.0, 1.0),
                      (region_4, 1.0, 1.0),
                      (region_5, 1.0, 1.0),
                      (1.0, 0.5, 0.0)),

             'green': ((0.0, 1.0, 1.0),
                       (region_2, 1.0, 1.0),
                       (region_3, 1.0, 1.0),
                       (region_4, 1.0, 1.0),
                       (region_5, 0.0, 0.0),
                       (1.0, 0.0, 0.0)),

             'blue':  ((0.0, 1.0, 1.0),
                       (region_2, 1.0, 1.0),
                       (region_3, 1.0, 1.0),
                       (region_4, 1.0, 1.0),
                       (region_5, 0.0, 0.0),
                       (1.0, 0.0, 0.0))
             }

    return LinearSegmentedColormap('RedWhite', cdict)


def blue_white_cmap(white_region=0.1):
    region_2 = (0.5 - white_region)/2.
    region_3 = 0.5 - white_region/2.
    region_4 = 0.5 + white_region/2.
    region_5 = 1. - (0.5 - white_region)/2.

    cdict = {'red':  ((0.0, 0.0, 0.0),
                      (region_2, 0.0, 0.0),
                      (region_3, 1.0, 1.0),
                      (region_4, 1.0, 1.0),
                      (region_5, 1.0, 1.0),
                      (1.0, 1.0, 1.0)),

             'green': ((0.0, 0.0, 0.0),
                       (region_2, 0.0, 0.0),
                       (region_3, 1.0, 1.0),
                       (region_4, 1.0, 1.0),
                       (region_5, 1.0, 1.0),
                       (1.0, 1.0, 1.0)),

             'blue':  ((0.0, 0.0, 0.5),
                       (region_2, 1.0, 1.0),
                       (region_3, 1.0, 1.0),
                       (region_4, 1.0, 1.0),
                       (region_5, 1.0, 1.0),
                       (1.0, 1.0, 1.0))
             }

    return LinearSegmentedColormap('BlueWhite', cdict)


def set_img_path(path):
    global id31_img_path
    id31_img_path = path


def set_cal_file(path):
    global id31_cal_file
    global id31_cal_matrix
    id31_cal_file = path
    id31_cal_matrix = np.transpose(np.load(id31_cal_file))


def get_filename(scan, img):
    filename = id31_img_path
    filename += str(scan).zfill(4)
    filename += '_'
    filename += str(img).zfill(4)
    filename += '.edf.gz'
    return filename

#def get_fast_filename(scan, img):
    

def get_p3_img(scan, img):
    imgdata = P3_Image(get_filename(scan, img))
    return imgdata


def get_p3_img_recal(scan, img):
    imgdata = P3_Image(get_filename(scan, img))
    imgdata.img = np.multiply(imgdata.img, id31_cal_matrix)
    return imgdata


def crop(img, centre, crop_size):
    cen_x = centre[0]
    cen_y = centre[1]
    crop_h = crop_size[0]
    crop_v = crop_size[1]
    return img[int(np.rint(cen_y-crop_v)):int(np.rint(cen_y+crop_v)),
               int(np.rint(cen_x-crop_h)):int(np.rint(cen_x+crop_h))]
